"""
Class for reading data from Neuralynx files.
This IO supports NCS, NEV, NSE and NTT file formats.


NCS contains the sampled signal for one channel
NEV contains events
NSE contains spikes and waveforms for mono electrodes
NTT contains spikes and waveforms for tetrodes

All Neuralynx files contain a 16 kilobyte text header followed by 0 or more fixed length records.
The format of the header has never been formally specified, however, the Neuralynx programs which
write them have followed a set of conventions which have varied over the years. Additionally,
other programs like Pegasus write files with somewhat varying headers. This variation requires
parsing to determine the exact version and type which is handled within this RawIO by the
NlxHeader class.

Ncs files contain a series of 1044 byte records, each of which contains 512 16 byte samples and
header information which includes a 64 bit timestamp in microseconds, a 16 bit channel number,
the sampling frequency in integral Hz, and the number of the 512 samples which are considered
valid samples (the remaining samples within the record are invalid). The Ncs file header usually
contains a specification of the sampling frequency, which may be rounded to an integral number
of Hz or may be fractional. The actual sampling frequency in terms of the underlying clock is
physically determined by the spacing of the timestamps between records.

These variations of header format and possible differences between the stated sampling frequency
and actual sampling frequency can create apparent time discrepancies in .Ncs files. Additionally,
the Neuralynx recording software can start and stop recording while continuing to write samples
to a single .Ncs file, which creates larger gaps in the time sequences of the samples.

This RawIO attempts to correct for these deviations where possible and present a single section of
contiguous samples with one sampling frequency, t_start, and length for each .Ncs file. These
sections are determined by the NcsSectionsFactory class. In the
event the gaps are larger, this RawIO only provides the samples from the first section as belonging
to one Segment.

This RawIO presents only a single Block and Segment.
:TODO: This should likely be changed to provide multiple segments and allow for
multiple .Ncs files in a directory with differing section structures.

Author: Julia Sprenger, Carlos Canova, Samuel Garcia, Peter N. Steinmetz.
"""


from ..baserawio import (BaseRawIO, _signal_channel_dtype, _signal_stream_dtype,
                _spike_channel_dtype, _event_channel_dtype)

import numpy as np
import os
import pathlib
import copy
from collections import (namedtuple, OrderedDict)

from neo.rawio.neuralynxrawio.ncssections import (NcsSection, NcsSectionsFactory)
from neo.rawio.neuralynxrawio.nlxheader import NlxHeader


class NeuralynxRawIO(BaseRawIO):
    """"
    Class for reading datasets recorded by Neuralynx.

    This version works with rawmode of one-dir for a single directory of files or one-file for
    a single file.

    Examples:
        >>> reader = NeuralynxRawIO(dirname='Cheetah_v5.5.1/original_data')
        >>> reader.parse_header()

            Inspect all files in the directory.

        >>> print(reader)

            Display all information about signal channels, units, segment size....
    """
    extensions = ['nse', 'ncs', 'nev', 'ntt']
    rawmode = 'one-dir'

    _ncs_dtype = [('timestamp', 'uint64'), ('channel_id', 'uint32'), ('sample_rate', 'uint32'),
                  ('nb_valid', 'uint32'), ('samples', 'int16', (NcsSection._RECORD_SIZE))]

    def __init__(self, dirname='', filename='', exclude_filename=None, keep_original_times=False,
                 **kargs):
        """
        Initialize io for either a directory of Ncs files or a single Ncs file.

        Parameters
        ----------
        dirname: str
            name of directory containing all files for dataset. If provided, filename is
            ignored.
        filename: str
            name of a single ncs, nse, nev, or ntt file to include in dataset. If used,
            dirname must not be provided.
        exclude_filename: str or list
            name of a single ncs, nse, nev or ntt file or list of such files. Expects plain
            filenames (without directory path).
        keep_original_times:
            if True, keep original start time as in files,
            otherwise set 0 of time to first time in dataset
        """
        if dirname != '':
            self.dirname = dirname
            self.rawmode = 'one-dir'
        elif filename != '':
            self.filename = filename
            self.rawmode = 'one-file'
        else:
            raise ValueError("One of dirname or filename must be provided.")

        self.keep_original_times = keep_original_times
        self.exclude_filename = exclude_filename
        BaseRawIO.__init__(self, **kargs)

    def _source_name(self):
        if self.rawmode == 'one-file':
            return self.filename
        else:
            return self.dirname

    # from memory_profiler import profile
    #
    # @profile()
    def _parse_header(self):

        stream_channels = []
        signal_channels = []
        spike_channels = []
        event_channels = []

        self.ncs_filenames = OrderedDict()  # (chan_name, chan_id): filename
        self.nse_ntt_filenames = OrderedDict()  # (chan_name, chan_id): filename
        self.nev_filenames = OrderedDict()  # chan_id: filename

        self._nev_memmap = {}
        self._spike_memmap = {}
        self.internal_unit_ids = []  # channel_index > ((channel_name, channel_id), unit_id)
        self.internal_event_ids = []
        self._empty_ncs = []  # this list contains filenames of empty files
        self._empty_nev = []
        self._empty_nse_ntt = []

        # Explore the directory looking for ncs, nev, nse and ntt
        # and construct channels headers.
        signal_annotations = []
        unit_annotations = []
        event_annotations = []

        if self.rawmode == 'one-dir':
            filenames = sorted(os.listdir(self.dirname))
            dirname = self.dirname
        else:
            if not os.path.isfile(self.filename):
                raise ValueError(f'Provided Filename is not a file: '
                                 f'{self.filename}. If you want to provide a '
                                 f'directory use the `dirname` keyword')

            dirname, fname = os.path.split(self.filename)
            filenames = [fname]

        if not isinstance(self.exclude_filename, (list, set, np.ndarray)):
            self.exclude_filename = [self.exclude_filename]

        # remove files that were explicitly excluded
        if self.exclude_filename is not None:
            for excl_file in self.exclude_filename:
                if excl_file in filenames:
                    filenames.remove(excl_file)

        for filename in filenames:
            filename = os.path.join(dirname, filename)

            _, ext = os.path.splitext(filename)
            ext = ext[1:]  # remove dot
            ext = ext.lower()  # make lower case for comparisons
            if ext not in self.extensions:
                continue

            # Skip Ncs files with only header. Other empty file types
            # will have an empty dataset constructed later.
            if (os.path.getsize(filename) <= NlxHeader.HEADER_SIZE) and ext in ['ncs']:
                self._empty_ncs.append(filename)
                continue

            # All file have more or less the same header structure
            info = NlxHeader(filename)
            chan_names = info['channel_names']
            chan_ids = info['channel_ids']

            for idx, chan_id in enumerate(chan_ids):
                chan_name = chan_names[idx]

                chan_uid = (chan_name, chan_id)
                if ext == 'ncs':
                    # a sampled signal channel
                    units = 'uV'
                    gain = info['bit_to_microVolt'][idx]
                    if info.get('input_inverted', False):
                        gain *= -1
                    offset = 0.
                    stream_id = 0
                    signal_channels.append((chan_name, str(chan_id), info['sampling_rate'],
                                         'int16', units, gain, offset, stream_id))
                    self.ncs_filenames[chan_uid] = filename
                    keys = [
                        'DspFilterDelay_µs',
                        'recording_opened',
                        'FileType',
                        'DspDelayCompensation',
                        'recording_closed',
                        'DspLowCutFilterType',
                        'HardwareSubSystemName',
                        'DspLowCutNumTaps',
                        'DSPLowCutFilterEnabled',
                        'HardwareSubSystemType',
                        'DspHighCutNumTaps',
                        'ADMaxValue',
                        'DspLowCutFrequency',
                        'DSPHighCutFilterEnabled',
                        'RecordSize',
                        'InputRange',
                        'DspHighCutFrequency',
                        'input_inverted',
                        'NumADChannels',
                        'DspHighCutFilterType',
                    ]
                    d = {k: info[k] for k in keys if k in info}
                    signal_annotations.append(d)

                elif ext in ('nse', 'ntt'):
                    # nse and ntt are pretty similar except for the waveform shape.
                    # A file can contain several unit_id (so several unit channel).
                    assert chan_id not in self.nse_ntt_filenames, \
                        'Several nse or ntt files have the same unit_id!!!'
                    self.nse_ntt_filenames[chan_uid] = filename

                    data = self._get_file_map(filename)
                    self._spike_memmap[chan_uid] = data

                    unit_ids = np.unique(data['unit_id'])
                    for unit_id in unit_ids:
                        # a spike channel for each (chan_id, unit_id)
                        self.internal_unit_ids.append((chan_uid, unit_id))

                        unit_name = "ch{}#{}#{}".format(chan_name, chan_id, unit_id)
                        unit_id = '{}'.format(unit_id)
                        wf_units = 'uV'
                        wf_gain = info['bit_to_microVolt'][idx]
                        if info.get('input_inverted', False):
                            wf_gain *= -1
                        wf_offset = 0.
                        wf_left_sweep = -1  # NOT KNOWN
                        wf_sampling_rate = info['sampling_rate']
                        spike_channels.append(
                            (unit_name, '{}'.format(unit_id), wf_units, wf_gain,
                             wf_offset, wf_left_sweep, wf_sampling_rate))
                        unit_annotations.append(dict(file_origin=filename))

                elif ext == 'nev':
                    # an event channel
                    # each ('event_id',  'ttl_input') give a new event channel
                    self.nev_filenames[chan_id] = filename

                    if os.path.getsize(filename) <= NlxHeader.HEADER_SIZE:
                        self._empty_nev.append(filename)
                        data = np.zeros((0,), dtype=nev_dtype)
                        internal_ids = []
                    else:
                        data = self._get_file_map(filename)
                        internal_ids = np.unique(data[['event_id', 'ttl_input']]).tolist()
                    for internal_event_id in internal_ids:
                        if internal_event_id not in self.internal_event_ids:
                            event_id, ttl_input = internal_event_id
                            name = '{} event_id={} ttl={}'.format(
                                chan_name, event_id, ttl_input)
                            event_channels.append((name, chan_id, 'event'))
                            self.internal_event_ids.append(internal_event_id)

                    self._nev_memmap[chan_id] = data

        signal_channels = np.array(signal_channels, dtype=_signal_channel_dtype)
        spike_channels = np.array(spike_channels, dtype=_spike_channel_dtype)
        event_channels = np.array(event_channels, dtype=_event_channel_dtype)

        # require all sampled signals, ncs files, to have the same sampling rate
        if signal_channels.size > 0:
            sampling_rate = np.unique(signal_channels['sampling_rate'])
            assert sampling_rate.size == 1
            self._sigs_sampling_rate = sampling_rate[0]
            signal_streams = [('signals', '0')]
        else:
            signal_streams = []
        signal_streams = np.array(signal_streams, dtype=_signal_stream_dtype)

        # set 2 attributes needed later for header in case there are no ncs files in dataset,
        #   e.g. Pegasus
        self._timestamp_limits = None
        self._nb_segment = 1

        # Read ncs files for gap detection and nb_segment computation.
        self._sigs_memmaps, ncsSegTimestampLimits = self.scan_ncs_files(self.ncs_filenames)
        if ncsSegTimestampLimits:
            self._ncs_seg_timestamp_limits = ncsSegTimestampLimits  # save copy
            self._nb_segment = ncsSegTimestampLimits.nb_segment
            self._sigs_length = ncsSegTimestampLimits.length.copy()
            self._timestamp_limits = ncsSegTimestampLimits.timestamp_limits.copy()
            self._sigs_t_start = ncsSegTimestampLimits.t_start.copy()
            self._sigs_t_stop = ncsSegTimestampLimits.t_stop.copy()

        # Determine timestamp limits in nev, nse file by scanning them.
        ts0, ts1 = None, None
        for _data_memmap in (self._spike_memmap, self._nev_memmap):
            for _, data in _data_memmap.items():
                ts = data['timestamp']
                if ts.size == 0:
                    continue
                if ts0 is None:
                    ts0 = ts[0]
                    ts1 = ts[-1]
                ts0 = min(ts0, ts[0])
                ts1 = max(ts1, ts[-1])

        # decide on segment and global start and stop times based on files available
        if self._timestamp_limits is None:
            # case  NO ncs but HAVE nev or nse
            self._timestamp_limits = [(ts0, ts1)]
            self._seg_t_starts = [ts0 / 1e6]
            self._seg_t_stops = [ts1 / 1e6]
            self.global_t_start = ts0 / 1e6
            self.global_t_stop = ts1 / 1e6
        elif ts0 is not None:
            # case  HAVE ncs AND HAVE nev or nse
            self.global_t_start = min(ts0 / 1e6, self._sigs_t_start[0])
            self.global_t_stop = max(ts1 / 1e6, self._sigs_t_stop[-1])
            self._seg_t_starts = list(self._sigs_t_start)
            self._seg_t_starts[0] = self.global_t_start
            self._seg_t_stops = list(self._sigs_t_stop)
            self._seg_t_stops[-1] = self.global_t_stop
        else:
            # case HAVE ncs but  NO nev or nse
            self._seg_t_starts = self._sigs_t_start
            self._seg_t_stops = self._sigs_t_stop
            self.global_t_start = self._sigs_t_start[0]
            self.global_t_stop = self._sigs_t_stop[-1]

        if self.keep_original_times:
            self.global_t_stop = self.global_t_stop - self.global_t_start
            self.global_t_start = 0

        # fill header dictionary
        self.header = {}
        self.header['nb_block'] = 1
        self.header['nb_segment'] = [self._nb_segment]
        self.header['signal_streams'] = signal_streams
        self.header['signal_channels'] = signal_channels
        self.header['spike_channels'] = spike_channels
        self.header['event_channels'] = event_channels

        # Annotations
        self._generate_minimal_annotations()
        bl_annotations = self.raw_annotations['blocks'][0]

        for seg_index in range(self._nb_segment):
            seg_annotations = bl_annotations['segments'][seg_index]

            for c in range(signal_streams.size):
                # one or no signal stream
                sig_ann = seg_annotations['signals'][c]
                # handle array annotations
                for key in signal_annotations[0].keys():
                    values = []
                    for c in range(signal_channels.size):
                        value = signal_annotations[0][key]
                        values.append(value)
                    values = np.array(values)
                    if values.ndim == 1:
                        # 'InputRange': is 2D and make bugs
                        sig_ann['__array_annotations__'][key] = values

            for c in range(spike_channels.size):
                unit_ann = seg_annotations['spikes'][c]
                unit_ann.update(unit_annotations[c])

            for c in range(event_channels.size):
                # annotations for channel events
                event_id, ttl_input = self.internal_event_ids[c]
                chan_id = event_channels[c]['id']

                ev_ann = seg_annotations['events'][c]
                ev_ann['file_origin'] = self.nev_filenames[chan_id]

                # ~ ev_ann['marker_id'] =
                # ~ ev_ann['nttl'] =
                # ~ ev_ann['digital_marker'] =
                # ~ ev_ann['analog_marker'] =

    def _get_file_map(self, filename):
        """
        Create memory maps when needed
        see also https://github.com/numpy/numpy/issues/19340
        """
        filename = pathlib.Path(filename)
        suffix = filename.suffix.lower()[1:]

        if suffix == 'ncs':
            return np.memmap(filename, dtype=self._ncs_dtype, mode='r',
                      offset=NlxHeader.HEADER_SIZE)

        elif suffix in ['nse', 'ntt']:
            info = NlxHeader(filename)
            dtype = get_nse_or_ntt_dtype(info, suffix)

            # return empty map if file does not contain data
            if os.path.getsize(filename) <= NlxHeader.HEADER_SIZE:
                self._empty_nse_ntt.append(filename)
                return np.zeros((0,), dtype=dtype)

            return np.memmap(filename, dtype=dtype, mode='r',
                             offset=NlxHeader.HEADER_SIZE)

        elif suffix == 'nev':
            return np.memmap(filename, dtype=nev_dtype, mode='r',
                             offset=NlxHeader.HEADER_SIZE)

        else:
            raise ValueError(f'Unknown file suffix {suffix}')

    # Accessors for segment times which are offset by appropriate global start time
    def _segment_t_start(self, block_index, seg_index):
        return self._seg_t_starts[seg_index] - self.global_t_start

    def _segment_t_stop(self, block_index, seg_index):
        return self._seg_t_stops[seg_index] - self.global_t_start

    def _get_signal_size(self, block_index, seg_index, stream_index):
        return self._sigs_length[seg_index]

    def _get_signal_t_start(self, block_index, seg_index, stream_index):
        return self._sigs_t_start[seg_index] - self.global_t_start

    def _get_analogsignal_chunk(self, block_index, seg_index, i_start, i_stop,
                                stream_index, channel_indexes):
        """
        Retrieve chunk of analog signal, a chunk being a set of contiguous samples.

        PARAMETERS
        ----------
        block_index:
            index of block in dataset, ignored as only 1 block in this implementation
        seg_index:
            index of segment to use
        i_start:
            sample index of first sample within segment to retrieve
        i_stop:
            sample index of last sample within segment to retrieve
        channel_indexes:
            list of channel indices to return data for

        RETURNS
        -------
            array of samples, with each requested channel in a column
        """
        if i_start is None:
            i_start = 0
        if i_stop is None:
            i_stop = self._sigs_length[seg_index]

        block_start = i_start // NcsSection._RECORD_SIZE
        block_stop = i_stop // NcsSection._RECORD_SIZE + 1
        sl0 = i_start % 512
        sl1 = sl0 + (i_stop - i_start)

        if channel_indexes is None:
            channel_indexes = slice(None)

        channel_ids = self.header['signal_channels'][channel_indexes]['id'].astype(int)
        channel_names = self.header['signal_channels'][channel_indexes]['name']

        # create buffer for samples
        sigs_chunk = np.zeros((i_stop - i_start, len(channel_ids)), dtype='int16')

        for i, chan_uid in enumerate(zip(channel_names, channel_ids)):
            data = self._sigs_memmaps[seg_index][chan_uid]
            sub = data[block_start:block_stop]
            sigs_chunk[:, i] = sub['samples'].flatten()[sl0:sl1]

        return sigs_chunk

    def _spike_count(self, block_index, seg_index, unit_index):
        chan_uid, unit_id = self.internal_unit_ids[unit_index]
        data = self._spike_memmap[chan_uid]
        ts = data['timestamp']

        ts0, ts1 = self._timestamp_limits[seg_index]

        # only count spikes inside the timestamp limits, inclusive, and for the specified unit
        keep = (ts >= ts0) & (ts <= ts1) & (unit_id == data['unit_id'])
        nb_spike = int(data[keep].size)
        return nb_spike

    def _get_spike_timestamps(self, block_index, seg_index, unit_index, t_start, t_stop):
        chan_uid, unit_id = self.internal_unit_ids[unit_index]
        data = self._spike_memmap[chan_uid]
        ts = data['timestamp']

        ts0, ts1 = self._timestamp_limits[seg_index]
        if t_start is not None:
            ts0 = int((t_start + self.global_t_start) * 1e6)
        if t_start is not None:
            ts1 = int((t_stop + self.global_t_start) * 1e6)

        keep = (ts >= ts0) & (ts <= ts1) & (unit_id == data['unit_id'])
        timestamps = ts[keep]
        return timestamps

    def _rescale_spike_timestamp(self, spike_timestamps, dtype):
        spike_times = spike_timestamps.astype(dtype)
        spike_times /= 1e6
        spike_times -= self.global_t_start
        return spike_times

    def _get_spike_raw_waveforms(self, block_index, seg_index, unit_index,
                                 t_start, t_stop):
        chan_uid, unit_id = self.internal_unit_ids[unit_index]
        data = self._spike_memmap[chan_uid]
        ts = data['timestamp']

        ts0, ts1 = self._timestamp_limits[seg_index]
        if t_start is not None:
            ts0 = int((t_start + self.global_t_start) * 1e6)
        if t_start is not None:
            ts1 = int((t_stop + self.global_t_start) * 1e6)

        keep = (ts >= ts0) & (ts <= ts1) & (unit_id == data['unit_id'])

        wfs = data[keep]['samples']
        if wfs.ndim == 2:
            # case for nse
            waveforms = wfs[:, None, :]
        else:
            # case for ntt change (n, 32, 4) to (n, 4, 32)
            waveforms = wfs.swapaxes(1, 2)

        return waveforms

    def _event_count(self, block_index, seg_index, event_channel_index):
        event_id, ttl_input = self.internal_event_ids[event_channel_index]
        chan_id = self.header['event_channels'][event_channel_index]['id']
        data = self._nev_memmap[chan_id]
        ts0, ts1 = self._timestamp_limits[seg_index]
        ts = data['timestamp']
        keep = (ts >= ts0) & (ts <= ts1) & (data['event_id'] == event_id) & \
               (data['ttl_input'] == ttl_input)
        nb_event = int(data[keep].size)
        return nb_event

    def _get_event_timestamps(self, block_index, seg_index, event_channel_index, t_start, t_stop):
        event_id, ttl_input = self.internal_event_ids[event_channel_index]
        chan_id = self.header['event_channels'][event_channel_index]['id']
        data = self._nev_memmap[chan_id]
        ts0, ts1 = self._timestamp_limits[seg_index]

        if t_start is not None:
            ts0 = int((t_start + self.global_t_start) * 1e6)
        if t_start is not None:
            ts1 = int((t_stop + self.global_t_start) * 1e6)

        ts = data['timestamp']
        keep = (ts >= ts0) & (ts <= ts1) & (data['event_id'] == event_id) & \
               (data['ttl_input'] == ttl_input)

        subdata = data[keep]
        timestamps = subdata['timestamp']
        labels = subdata['event_string'].astype('U')
        durations = None
        return timestamps, durations, labels

    def _rescale_event_timestamp(self, event_timestamps, dtype, event_channel_index):
        event_times = event_timestamps.astype(dtype)
        event_times /= 1e6
        event_times -= self.global_t_start
        return event_times

    def scan_ncs_files(self, ncs_filenames):
        """
        Given a list of ncs files, read their basic structure.

        PARAMETERS:
        ------
        ncs_filenames - list of ncs filenames to scan.

        RETURNS:
        ------
        memmaps
            [ {} for seg_index in range(self._nb_segment) ][chan_uid]
        seg_time_limits
            SegmentTimeLimits for sections in scanned Ncs files

        Files will be scanned to determine the sections of records. If file is a single
        section of records, this scan is brief, otherwise it will check each record which may
        take some time.
        """

        # :TODO: Needs to account for gaps and start and end times potentially
        #    being different in different groups of channels. These groups typically
        #    correspond to the channels collected by a single ADC card.
        if len(ncs_filenames) == 0:
            return None, None

        # Build dictionary of chan_uid to associated NcsSections, memmap and NlxHeaders. Only
        # construct new NcsSections when it is different from that for the preceding file.
        chanSectMap = dict()
        for chan_uid, ncs_filename in self.ncs_filenames.items():

            data = self._get_file_map(ncs_filename)
            nlxHeader = NlxHeader(ncs_filename)

            if not chanSectMap or (chanSectMap and
                    not NcsSectionsFactory._verifySectionsStructure(data,
                    lastNcsSections)):
                lastNcsSections = NcsSectionsFactory.build_for_ncs_file(data, nlxHeader)
            chanSectMap[chan_uid] = [lastNcsSections, nlxHeader, ncs_filename]
            del data

        # Construct an inverse dictionary from NcsSections to list of associated chan_uids
        revSectMap = dict()
        for k, v in chanSectMap.items():
            revSectMap.setdefault(v[0], []).append(k)

        # If there is only one NcsSections structure in the set of ncs files, there should only
        # be one entry. Otherwise this is presently unsupported.
        if len(revSectMap) > 1:
            raise IOError(f'ncs files have {len(revSectMap)} different sections '
                          f'structures. Unsupported configuration.')

        seg_time_limits = SegmentTimeLimits(nb_segment=len(lastNcsSections.sects),
                                            t_start=[], t_stop=[], length=[],
                                            timestamp_limits=[])
        memmaps = [{} for seg_index in range(seg_time_limits.nb_segment)]

        # create segment with subdata block/t_start/t_stop/length for each channel
        for i, fileEntry in enumerate(self.ncs_filenames.items()):
            chan_uid = fileEntry[0]
            data = self._get_file_map(chanSectMap[chan_uid][2])

            # create a memmap for each record section of the current file
            curSects = chanSectMap[chan_uid][0]
            for seg_index in range(len(curSects.sects)):

                curSect = curSects.sects[seg_index]
                subdata = data[curSect.startRec:(curSect.endRec + 1)]
                memmaps[seg_index][chan_uid] = subdata

                # create segment timestamp limits based on only NcsSections structure in use
                if i == 0:
                    numSampsLastSect = subdata[-1]['nb_valid']
                    ts0 = subdata[0]['timestamp']
                    ts1 = NcsSectionsFactory.calc_sample_time(curSects.sampFreqUsed,
                                                              subdata[-1]['timestamp'],
                                                              numSampsLastSect)
                    seg_time_limits.timestamp_limits.append((ts0, ts1))
                    t_start = ts0 / 1e6
                    seg_time_limits.t_start.append(t_start)
                    t_stop = ts1 / 1e6
                    seg_time_limits.t_stop.append(t_stop)
                    # :NOTE: This should really be the total of nb_valid in records, but this
                    #  allows the last record of a section to be shorter, the most common case.
                    #  Have never seen a section of records with not full records before the last.
                    length = (subdata.size - 1) * NcsSection._RECORD_SIZE + numSampsLastSect
                    seg_time_limits.length.append(length)

        return memmaps, seg_time_limits


# time limits for set of segments
SegmentTimeLimits = namedtuple("SegmentTimeLimits", ['nb_segment', 't_start', 't_stop', 'length',
                                                     'timestamp_limits'])


nev_dtype = [
    ('reserved', '<i2'),
    ('system_id', '<i2'),
    ('data_size', '<i2'),
    ('timestamp', '<u8'),
    ('event_id', '<i2'),
    ('ttl_input', '<i2'),
    ('crc_check', '<i2'),
    ('dummy1', '<i2'),
    ('dummy2', '<i2'),
    ('extra', '<i4', (8,)),
    ('event_string', 'S128'),
]


def get_nse_or_ntt_dtype(info, ext):
    """
    For NSE and NTT the dtype depend on the header.

    """
    dtype = [('timestamp', 'uint64'), ('channel_id', 'uint32'), ('unit_id', 'uint32')]

    # count feature
    nb_feature = 0
    for k in info.keys():
        if k.startswith('Feature '):
            nb_feature += 1
    dtype += [('features', 'int32', (nb_feature,))]

    # count sample
    if ext == 'nse':
        nb_sample = info['WaveformLength']
        dtype += [('samples', 'int16', (nb_sample,))]
    elif ext == 'ntt':
        nb_sample = info['WaveformLength']
        nb_chan = 4  # check this if not tetrode
        dtype += [('samples', 'int16', (nb_sample, nb_chan))]

    return dtype



"""
Tests of neo.io.BiocamIO
"""

import unittest

from neo.io import BiocamIO
from neo.test.iotest.common_io_test import BaseTestIO


class TestBiocamIO(BaseTestIO, unittest.TestCase, ):
    ioclass = BiocamIO
    entities_to_download = [
        'biocam'
    ]
    entities_to_test = [
        'biocam/biocam_hw3.0_fw1.6.brw'
    ]


if __name__ == "__main__":
    unittest.main()



import unittest

from neo.rawio.winwcprawio import WinWcpRawIO
from neo.test.rawiotest.common_rawio_test import BaseTestRawIO


class TestWinWcpRawIO(BaseTestRawIO, unittest.TestCase, ):
    rawioclass = WinWcpRawIO
    entities_to_download = [
        'winwcp'
    ]
    entities_to_test = [
        'winwcp/File_winwcp_1.wcp'
    ]


if __name__ == "__main__":
    unittest.main()



import datetime
import distutils.version
import os
import re
from collections import OrderedDict


class NlxHeader(OrderedDict):
    """
    Representation of basic information in all 16 kbytes Neuralynx file headers,
    including dates opened and closed if given.
    """

    HEADER_SIZE = 2 ** 14  # Neuralynx files have a txt header of 16kB

    # helper function to interpret boolean keys
    def _to_bool(txt):
        if txt == 'True':
            return True
        elif txt == 'False':
            return False
        else:
            raise Exception('Can not convert %s to bool' % txt)

    # keys that may be present in header which we parse
    txt_header_keys = [
        ('AcqEntName', 'channel_names', None),  # used
        ('FileType', '', None),
        ('FileVersion', '', None),
        ('RecordSize', '', None),
        ('HardwareSubSystemName', '', None),
        ('HardwareSubSystemType', '', None),
        ('SamplingFrequency', 'sampling_rate', float),  # used
        ('ADMaxValue', '', None),
        ('ADBitVolts', 'bit_to_microVolt', None),  # used
        ('NumADChannels', '', None),
        ('ADChannel', 'channel_ids', None),  # used
        ('InputRange', '', None),
        ('InputInverted', 'input_inverted', _to_bool),  # used
        ('DSPLowCutFilterEnabled', '', None),
        ('DspLowCutFrequency', '', None),
        ('DspLowCutNumTaps', '', None),
        ('DspLowCutFilterType', '', None),
        ('DSPHighCutFilterEnabled', '', None),
        ('DspHighCutFrequency', '', None),
        ('DspHighCutNumTaps', '', None),
        ('DspHighCutFilterType', '', None),
        ('DspDelayCompensation', '', None),
        ('DspFilterDelay_µs', '', None),
        ('DisabledSubChannels', '', None),
        ('WaveformLength', '', int),
        ('AlignmentPt', '', None),
        ('ThreshVal', '', None),
        ('MinRetriggerSamples', '', None),
        ('SpikeRetriggerTime', '', None),
        ('DualThresholding', '', None),
        (r'Feature \w+ \d+', '', None),
        ('SessionUUID', '', None),
        ('FileUUID', '', None),
        ('CheetahRev', '', None),  # only for older versions of Cheetah
        ('ProbeName', '', None),
        ('OriginalFileName', '', None),
        ('TimeCreated', '', None),
        ('TimeClosed', '', None),
        ('ApplicationName', '', None),  # also include version number when present
        ('AcquisitionSystem', '', None),
        ('ReferenceChannel', '', None),
        ('NLX_Base_Class_Type', '', None)  # in version 4 and earlier versions of Cheetah
    ]

    # Filename and datetime may appear in header lines starting with # at
    # beginning of header or in later versions as a property. The exact format
    # used depends on the application name and its version as well as the
    # -FileVersion property.
    #
    # There are 4 styles understood by this code and the patterns used for parsing
    # the items within each are stored in a dictionary. Each dictionary is then
    # stored in main dictionary keyed by an abbreviation for the style.
    header_pattern_dicts = {
        # BML
        'bml': dict(
            datetime1_regex=r'## Time Opened: \(m/d/y\): (?P<date>\S+)'
                            r'  At Time: (?P<time>\S+)',
            filename_regex=r'## File Name: (?P<filename>\S+)',
            datetimeformat='%m/%d/%y %H:%M:%S.%f'),
        # Cheetah after version 1 and before version 5
        'bv5': dict(
            datetime1_regex=r'## Time Opened: \(m/d/y\): (?P<date>\S+)'
                            r'  At Time: (?P<time>\S+)',
            filename_regex=r'## File Name: (?P<filename>\S+)',
            datetimeformat='%m/%d/%Y %H:%M:%S.%f'),
        # Cheetah version 5.4.0
        'v5.4.0': dict(
            datetime1_regex=r'## Time Opened \(m/d/y\): (?P<date>\S+)'
                            r'  At Time: (?P<time>\S+)',
            datetime2_regex=r'## Time Closed \(m/d/y\): (?P<date>\S+)'
                            r'  At Time: (?P<time>\S+)',
            filename_regex=r'## File Name: (?P<filename>\S+)',
            datetimeformat='%m/%d/%Y %H:%M:%S.%f'),
        # Cheetah version 5 before and including v 5.6.4 as well as version 1
        'bv5.6.4': dict(
            datetime1_regex=r'## Time Opened \(m/d/y\): (?P<date>\S+)'
                            r'  \(h:m:s\.ms\) (?P<time>\S+)',
            datetime2_regex=r'## Time Closed \(m/d/y\): (?P<date>\S+)'
                            r'  \(h:m:s\.ms\) (?P<time>\S+)',
            filename_regex=r'## File Name (?P<filename>\S+)',
            datetimeformat='%m/%d/%Y %H:%M:%S.%f'),
        'neuraview2': dict(
            datetime1_regex=r'## Date Opened: \(mm/dd/yyy\): (?P<date>\S+)'
                            r' At Time: (?P<time>\S+)',
            datetime2_regex=r'## Date Closed: \(mm/dd/yyy\): (?P<date>\S+)'
                            r' At Time: (?P<time>\S+)',
            filename_regex=r'## File Name: (?P<filename>\S+)',
            datetimeformat='%m/%d/%Y %H:%M:%S'),
        # Cheetah after v 5.6.4 and default for others such as Pegasus
        'def': dict(
            datetime1_regex=r'-TimeCreated (?P<date>\S+) (?P<time>\S+)',
            datetime2_regex=r'-TimeClosed (?P<date>\S+) (?P<time>\S+)',
            filename_regex=r'-OriginalFileName "?(?P<filename>\S+)"?',
            datetimeformat='%Y/%m/%d %H:%M:%S')
    }

    def __init__(self, filename):
        """
        Factory function to build NlxHeader for a given file.
        """
        super(OrderedDict, self).__init__()
        with open(filename, 'rb') as f:
            txt_header = f.read(NlxHeader.HEADER_SIZE)
        txt_header = txt_header.strip(b'\x00').decode('latin-1')

        # must start with 8 # characters
        assert txt_header.startswith("########"),\
            'Neuralynx files must start with 8 # characters.'

        # find keys
        for k1, k2, type_ in NlxHeader.txt_header_keys:
            pattern = r'-(?P<name>' + k1 + r')\s+(?P<value>[\S ]*)'
            matches = re.findall(pattern, txt_header)
            for match in matches:
                if k2 == '':
                    name = match[0]
                else:
                    name = k2
                value = match[1].rstrip(' ')
                if type_ is not None:
                    value = type_(value)
                self[name] = value

        # if channel_ids or s not in self then the filename is used
        name = os.path.splitext(os.path.basename(filename))[0]

        # convert channel ids
        if 'channel_ids' in self:
            chid_entries = re.findall(r'\w+', self['channel_ids'])
            self['channel_ids'] = [int(c) for c in chid_entries]
        else:
            self['channel_ids'] = ['unknown']

        # convert channel names
        if 'channel_names' in self:
            name_entries = re.findall(r'\w+', self['channel_names'])
            if len(name_entries) == 1:
                self['channel_names'] = name_entries * len(self['channel_ids'])
            assert len(self['channel_names']) == len(self['channel_ids']), \
                'Number of channel ids does not match channel names.'
        else:
            self['channel_names'] = ['unknown'] * len(self['channel_ids'])

        # version and application name
        # older Cheetah versions with CheetahRev property
        if 'CheetahRev' in self:
            assert 'ApplicationName' not in self
            self['ApplicationName'] = 'Cheetah'
            app_version = self['CheetahRev']
        # new file version 3.4 does not contain CheetahRev property, but ApplicationName instead
        elif 'ApplicationName' in self:
            pattern = r'(\S*) "([\S ]*)"'
            match = re.findall(pattern, self['ApplicationName'])
            assert len(match) == 1, 'impossible to find application name and version'
            self['ApplicationName'], app_version = match[0]
        # BML Ncs file contain neither property, but 'NLX_Base_Class_Type'
        elif 'NLX_Base_Class_Type' in txt_header:
            self['ApplicationName'] = 'BML'
            app_version = "2.0"
        # Neuraview Ncs file contained neither property nor
        # NLX_Base_Class_Type information
        else:
            self['ApplicationName'] = 'Neuraview'
            app_version = '2'

        self['ApplicationVersion'] = distutils.version.LooseVersion(app_version)

        # convert bit_to_microvolt
        if 'bit_to_microVolt' in self:
            btm_entries = re.findall(r'\S+', self['bit_to_microVolt'])
            if len(btm_entries) == 1:
                btm_entries = btm_entries * len(self['channel_ids'])
            self['bit_to_microVolt'] = [float(e) * 1e6 for e in btm_entries]
            assert len(self['bit_to_microVolt']) == len(self['channel_ids']), \
                'Number of channel ids does not match bit_to_microVolt conversion factors.'

        if 'InputRange' in self:
            ir_entries = re.findall(r'\w+', self['InputRange'])
            if len(ir_entries) == 1:
                self['InputRange'] = [int(ir_entries[0])] * len(chid_entries)
            else:
                self['InputRange'] = [int(e) for e in ir_entries]
            assert len(self['InputRange']) == len(chid_entries), \
                'Number of channel ids does not match input range values.'

        # Format of datetime depends on app name, app version
        # :TODO: this works for current examples but is not likely actually related
        # to app version in this manner.
        an = self['ApplicationName']
        if an == 'Cheetah':
            av = self['ApplicationVersion']
            if av <= '2':  # version 1 uses same as older versions
                hpd = NlxHeader.header_pattern_dicts['bv5.6.4']
            elif av < '5':
                hpd = NlxHeader.header_pattern_dicts['bv5']
            elif av <= '5.4.0':
                hpd = NlxHeader.header_pattern_dicts['v5.4.0']
            elif av <= '5.6.4':
                hpd = NlxHeader.header_pattern_dicts['bv5.6.4']
            else:
                hpd = NlxHeader.header_pattern_dicts['def']
        elif an == 'BML':
            hpd = NlxHeader.header_pattern_dicts['bml']
            av = "2"
        elif an == 'Neuraview':
            hpd = NlxHeader.header_pattern_dicts['neuraview2']
            av = "2"
        else:
            an = "Unknown"
            av = "NA"
            hpd = NlxHeader.header_pattern_dicts['def']

        # opening time
        sr = re.search(hpd['datetime1_regex'], txt_header)
        if not sr:
            raise IOError("No matching header open date/time for application {} " +
                          "version {}. Please contact developers.".format(an, av))
        else:
            dt1 = sr.groupdict()
            self['recording_opened'] = datetime.datetime.strptime(
                dt1['date'] + ' ' + dt1['time'], hpd['datetimeformat'])

        # close time, if available
        if 'datetime2_regex' in hpd:
            sr = re.search(hpd['datetime2_regex'], txt_header)
            if not sr:
                raise IOError("No matching header close date/time for application {} " +
                              "version {}. Please contact developers.".format(an, av))
            else:
                dt2 = sr.groupdict()
                self['recording_closed'] = datetime.datetime.strptime(
                    dt2['date'] + ' ' + dt2['time'], hpd['datetimeformat'])

    def type_of_recording(self):
        """
        Determines type of recording in Ncs file with this header.

        RETURN:
            one of 'PRE4','BML','DIGITALLYNX','DIGITALLYNXSX','UNKNOWN'
        """

        if 'NLX_Base_Class_Type' in self:

            # older style standard neuralynx acquisition with rounded sampling frequency
            if self['NLX_Base_Class_Type'] == 'CscAcqEnt':
                return 'PRE4'

            # BML style with fractional frequency and microsPerSamp
            elif self['NLX_Base_Class_Type'] == 'BmlAcq':
                return 'BML'

            else:
                return 'UNKNOWN'

        elif 'HardwareSubSystemType' in self:

            # DigitalLynx
            if self['HardwareSubSystemType'] == 'DigitalLynx':
                return 'DIGITALLYNX'

            # DigitalLynxSX
            elif self['HardwareSubSystemType'] == 'DigitalLynxSX':
                return 'DIGITALLYNXSX'

        elif 'FileType' in self:

            if self['FileVersion'] in ['3.3', '3.4']:
                return self['AcquisitionSystem'].split()[1].upper()

            else:
                return 'UNKNOWN'

        else:
            return 'UNKNOWN'



"""
Class for reading data from a 3-brain Biocam system.

See:
https://www.3brain.com/products/single-well/biocam-x

Author : Alessio Buccino
"""

from .baserawio import (BaseRawIO, _signal_channel_dtype, _signal_stream_dtype,
                        _spike_channel_dtype, _event_channel_dtype)

import numpy as np

try:
    import h5py
    HAVE_H5PY = True
except ImportError:
    HAVE_H5PY = False


class BiocamRawIO(BaseRawIO):
    """
    Class for reading data from a Biocam h5 file.

    Usage:
        >>> import neo.rawio
        >>> r = neo.rawio.BiocamRawIO(filename='biocam.h5')
        >>> r.parse_header()
        >>> print(r)
        >>> raw_chunk = r.get_analogsignal_chunk(block_index=0, seg_index=0,
                                                 i_start=0, i_stop=1024,  
                                                 channel_names=channel_names)
        >>> float_chunk = r.rescale_signal_raw_to_float(raw_chunk, dtype='float64',
                                                        channel_indexes=[0, 3, 6])
    """
    extensions = ['h5']
    rawmode = 'one-file'

    def __init__(self, filename=''):
        BaseRawIO.__init__(self)
        self.filename = filename

    def _source_name(self):
        return self.filename

    def _parse_header(self):
        assert HAVE_H5PY, 'h5py is not installed'
        self._header_dict = open_biocam_file_header(self.filename)
        self._num_channels = self._header_dict["num_channels"]
        self._num_frames = self._header_dict["num_frames"]
        self._sampling_rate = self._header_dict["sampling_rate"]
        self._filehandle = self._header_dict["file_handle"]
        self._read_function = self._header_dict["read_function"]
        self._channels = self._header_dict["channels"]
        gain = self._header_dict["gain"]
        offset = self._header_dict["offset"]

        signal_streams = np.array([('Signals', '0')], dtype=_signal_stream_dtype)

        sig_channels = []
        for c, chan in enumerate(self._channels):
            ch_name = f'ch{chan[0]}-{chan[1]}'
            chan_id = str(c + 1)
            sr = self._sampling_rate  # Hz
            dtype = "uint16"
            units = 'uV'
            gain = gain
            offset = offset
            stream_id = '0'
            sig_channels.append((ch_name, chan_id, sr, dtype, units, gain, offset, stream_id))
        sig_channels = np.array(sig_channels, dtype=_signal_channel_dtype)

        # No events
        event_channels = []
        event_channels = np.array(event_channels, dtype=_event_channel_dtype)

        # No spikes
        spike_channels = []
        spike_channels = np.array(spike_channels, dtype=_spike_channel_dtype)

        self.header = {}
        self.header['nb_block'] = 1
        self.header['nb_segment'] = [1]
        self.header['signal_streams'] = signal_streams
        self.header['signal_channels'] = sig_channels
        self.header['spike_channels'] = spike_channels
        self.header['event_channels'] = event_channels

        self._generate_minimal_annotations()

    def _segment_t_start(self, block_index, seg_index):
        all_starts = [[0.]]
        return all_starts[block_index][seg_index]

    def _segment_t_stop(self, block_index, seg_index):
        t_stop = self._num_frames / self._sampling_rate
        all_stops = [[t_stop]]
        return all_stops[block_index][seg_index]

    def _get_signal_size(self, block_index, seg_index, stream_index):
        assert stream_index == 0
        return self._num_frames

    def _get_signal_t_start(self, block_index, seg_index, stream_index):
        assert stream_index == 0
        return self._segment_t_start(block_index, seg_index)

    def _get_analogsignal_chunk(self, block_index, seg_index, i_start, i_stop,
                                stream_index, channel_indexes):
        if i_start is None:
            i_start = 0
        if i_stop is None:
            i_stop = self._num_frames
        if channel_indexes is None:
            channel_indexes = slice(None)
        data = self._read_function(self._filehandle, i_start, i_stop, self._num_channels)
        return data[:, channel_indexes]


def open_biocam_file_header(filename):
    """Open a Biocam hdf5 file, read and return the recording info, pick te correct method to access raw data,
    and return this to the caller."""
    assert HAVE_H5PY, 'h5py is not installed'

    rf = h5py.File(filename, 'r')
    # Read recording variables
    rec_vars = rf.require_group('3BRecInfo/3BRecVars/')
    bit_depth = rec_vars['BitDepth'][0]
    max_uv = rec_vars['MaxVolt'][0]
    min_uv = rec_vars['MinVolt'][0]
    n_frames = rec_vars['NRecFrames'][0]
    sampling_rate = rec_vars['SamplingRate'][0]
    signal_inv = rec_vars['SignalInversion'][0]

    # Get the actual number of channels used in the recording
    file_format = rf['3BData'].attrs.get('Version', None)
    format_100 = False
    if file_format == 100:
        n_channels = len(rf['3BData/Raw'][0])
        format_100 = True
    elif file_format in (101, 102) or file_format is None:
        n_channels = int(rf['3BData/Raw'].shape[0] / n_frames)
    else:
        raise Exception('Unknown data file format.')

    # # get channels
    channels = rf['3BRecInfo/3BMeaStreams/Raw/Chs'][:]

    # determine correct function to read data
    if format_100:
        if signal_inv == 1:
            read_function = readHDF5t_100
        elif signal_inv == 1:
            read_function = readHDF5t_100_i
        else:
            raise Exception("Unknown signal inversion")
    else:
        if signal_inv == 1:
            read_function = readHDF5t_101
        elif signal_inv == 1:
            read_function = readHDF5t_101_i
        else:
            raise Exception("Unknown signal inversion")

    gain = (max_uv - min_uv) / (2 ** bit_depth)
    offset = min_uv

    return dict(file_handle=rf, num_frames=n_frames, sampling_rate=sampling_rate, num_channels=n_channels,
                channels=channels, file_format=file_format, signal_inv=signal_inv,
                read_function=read_function, gain=gain, offset=offset)


def readHDF5t_100(rf, t0, t1, nch):
    return rf['3BData/Raw'][t0:t1]


def readHDF5t_100_i(rf, t0, t1, nch):
    return 4096 - rf['3BData/Raw'][t0:t1]


def readHDF5t_101(rf, t0, t1, nch):
    return rf['3BData/Raw'][nch * t0:nch * t1].reshape((t1 - t0, nch), order='C')


def readHDF5t_101_i(rf, t0, t1, nch):
    return 4096 - rf['3BData/Raw'][nch * t0:nch * t1].reshape((t1 - t0, nch), order='C')



"""
Class for reading data from maxwell biosystem device:
  * MaxOne
  * MaxTwo

https://www.mxwbio.com/resources/mea/

The implementation is a mix between:
  * the implementation in spikeextractors
     https://github.com/SpikeInterface/spikeextractors/blob/master/spikeextractors/extractors/maxwellextractors/maxwellextractors.py
 * the implementation in spyking-circus
    https://github.com/spyking-circus/spyking-circus/blob/master/circus/files/maxwell.py

The implementation do not handle spike at the moment.

For maxtwo device, each well will be a different signal stream.

Author : Samuel Garcia, Alessio Buccino, Pierre Yger
"""
import os
from pathlib import Path
import platform
from urllib.request import urlopen

from .baserawio import (BaseRawIO, _signal_channel_dtype, _signal_stream_dtype,
                        _spike_channel_dtype, _event_channel_dtype)

import numpy as np

try:
    import h5py
    HAVE_H5 = True
except ImportError:
    HAVE_H5 = False


class MaxwellRawIO(BaseRawIO):
    """
    Class for reading MaxOne or MaxTwo files.
    """
    extensions = ['h5']
    rawmode = 'one-file'

    def __init__(self, filename='', rec_name=None):
        BaseRawIO.__init__(self)
        self.filename = filename
        self.rec_name = rec_name

    def _source_name(self):
        return self.filename

    def _parse_header(self):
        try:
            import MEArec as mr
            HAVE_MEAREC = True
        except ImportError:
            HAVE_MEAREC = False
        assert HAVE_H5, 'h5py is not installed'

        h5 = h5py.File(self.filename, mode='r')
        self.h5_file = h5
        version = h5['version'][0].decode()

        # create signal stream
        # one stream per well
        signal_streams = []
        if int(version) == 20160704:
            self._old_format = True
            signal_streams.append(('well000', 'well000'))
        elif int(version) > 20160704:
            # multi stream stream (one well is one stream)
            self._old_format = False
            stream_ids = list(h5['wells'].keys())
            for stream_id in stream_ids:
                rec_names = list(h5['wells'][stream_id].keys())
                if len(rec_names) > 1:
                    if self.rec_name is None:
                        raise ValueError("Detected multiple recordings. Please select a "
                                         "single recording using the `rec_name` paramter. "
                                         f"Possible rec_name {rec_names}")
                else:
                    self.rec_name = rec_names[0]
                signal_streams.append((stream_id, stream_id))
        else:
            raise NotImplementedError(
                f'This version {version} is not supported')
        signal_streams = np.array(signal_streams, dtype=_signal_stream_dtype)

        # create signal channels
        max_sig_length = 0
        self._signals = {}
        sig_channels = []
        for stream_id in signal_streams['id']:
            if int(version) == 20160704:
                sr = 20000.
                settings = h5["settings"]
                if 'lsb' in settings:
                    gain_uV = settings['lsb'][0] * 1e6
                else:
                    if "gain" not in settings:
                        print("'gain' amd 'lsb' not found in settings. "
                              "Setting gain to 512 (default)")
                        gain = 512
                    else:
                        gain = settings['gain'][0]
                    gain_uV = 3.3 / (1024 * gain) * 1e6
                sigs = h5['sig']
                mapping = h5["mapping"]
                ids = np.array(mapping['channel'])
                ids = ids[ids >= 0]
                self._channel_slice = ids
            elif int(version) > 20160704:
                settings = h5['wells'][stream_id][self.rec_name]['settings']
                sr = settings['sampling'][0]
                if 'lsb' in settings:
                    gain_uV = settings['lsb'][0] * 1e6
                else:
                    if "gain" not in settings:
                        print("'gain' amd 'lsb' not found in settings. "
                              "Setting gain to 512 (default)")
                        gain = 512
                    else:
                        gain = settings['gain'][0]
                    gain_uV = 3.3 / (1024 * gain) * 1e6
                mapping = settings['mapping']
                sigs = h5['wells'][stream_id][self.rec_name]['groups']['routed']['raw']

            channel_ids = np.array(mapping['channel'])
            electrode_ids = np.array(mapping['electrode'])
            mask = channel_ids >= 0
            channel_ids = channel_ids[mask]
            electrode_ids = electrode_ids[mask]

            for i, chan_id in enumerate(channel_ids):
                elec_id = electrode_ids[i]
                ch_name = f'ch{chan_id} elec{elec_id}'
                offset_uV = 0
                sig_channels.append((ch_name, str(chan_id), sr, 'uint16', 'uV',
                                     gain_uV, offset_uV, stream_id))

            self._signals[stream_id] = sigs
            max_sig_length = max(max_sig_length, sigs.shape[1])

        self._t_stop = max_sig_length / sr

        sig_channels = np.array(sig_channels, dtype=_signal_channel_dtype)

        spike_channels = []
        spike_channels = np.array(spike_channels, dtype=_spike_channel_dtype)

        event_channels = []
        event_channels = np.array(event_channels, dtype=_event_channel_dtype)

        self.header = {}
        self.header['nb_block'] = 1
        self.header['nb_segment'] = [1]
        self.header['signal_streams'] = signal_streams
        self.header['signal_channels'] = sig_channels
        self.header['spike_channels'] = spike_channels
        self.header['event_channels'] = event_channels

        self._generate_minimal_annotations()
        bl_ann = self.raw_annotations['blocks'][0]
        bl_ann['maxwell_version'] = version

    def _segment_t_start(self, block_index, seg_index):
        return 0.

    def _segment_t_stop(self, block_index, seg_index):
        return self._t_stop

    def _get_signal_size(self, block_index, seg_index, stream_index):
        stream_id = self.header['signal_streams'][stream_index]['id']
        sigs = self._signals[stream_id]
        return sigs.shape[1]

    def _get_signal_t_start(self, block_index, seg_index, stream_index):
        return 0.

    def _get_analogsignal_chunk(self, block_index, seg_index, i_start, i_stop,
                                stream_index, channel_indexes):
        stream_id = self.header['signal_streams'][stream_index]['id']
        sigs = self._signals[stream_id]

        if i_start is None:
            i_start = 0
        if i_stop is None:
            i_stop = sigs.shape[1]

        resorted_indexes = None
        if channel_indexes is None:
            channel_indexes = slice(None)
        else:
            if np.array(channel_indexes).size > 1 and np.any(np.diff(channel_indexes) < 0):
                # get around h5py constraint that it does not allow datasets
                # to be indexed out of order
                sorted_channel_indexes = np.sort(channel_indexes)
                resorted_indexes = np.array(
                    [list(channel_indexes).index(ch) for ch in sorted_channel_indexes])

        try:
            if resorted_indexes is None:
                if self._old_format:
                    sigs = sigs[self._channel_slice, i_start:i_stop]
                    sigs = sigs[channel_indexes]
                else:
                    sigs = sigs[channel_indexes, i_start:i_stop]
            else:
                if self._old_format:
                    sigs = sigs[self._channel_slice, i_start:i_stop]
                    sigs = sigs[sorted_channel_indexes]
                else:
                    sigs = sigs[sorted_channel_indexes, i_start:i_stop]
                sigs = sigs[resorted_indexes]
        except OSError as e:
            print('*' * 10)
            print(_hdf_maxwell_error)
            print('*' * 10)
            raise(e)
        sigs = sigs.T

        return sigs


_hdf_maxwell_error = """Maxwell file format is based on HDF5.
The internal compression requires a custom plugin!!!
This is a big pain for the end user.
You, as a end user, should ask Maxwell company to change this.
Please visit this page and install the missing decompression libraries:
https://share.mxwbio.com/d/4742248b2e674a85be97/
Then, link the decompression library by setting the `HDF5_PLUGIN_PATH` to your 
installation location, e.g. via
os.environ['HDF5_PLUGIN_PATH'] = '/path/to/cutum/hdf5/plugin/'

Alternatively, you can use the auto_install_maxwell_hdf5_compression_plugin() below
function that do it automagically.
"""


def auto_install_maxwell_hdf5_compression_plugin(hdf5_plugin_path=None, force_download=True):
    if hdf5_plugin_path is None:
        hdf5_plugin_path = os.getenv('HDF5_PLUGIN_PATH', None)
        if hdf5_plugin_path is None:
            hdf5_plugin_path = Path.home() / 'hdf5_plugin_path_maxwell'
            os.environ['HDF5_PLUGIN_PATH'] = str(hdf5_plugin_path)
    hdf5_plugin_path = Path(hdf5_plugin_path)
    hdf5_plugin_path.mkdir(exist_ok=True)

    if platform.system() == 'Linux':
        remote_lib = 'https://share.mxwbio.com/d/4742248b2e674a85be97/files/?p=%2FLinux%2Flibcompression.so&dl=1'
        local_lib = hdf5_plugin_path / 'libcompression.so'
    elif platform.system() == 'Darwin':
        remote_lib = 'https://share.mxwbio.com/d/4742248b2e674a85be97/files/?p=%2FMacOS%2Flibcompression.dylib&dl=1'
        local_lib = hdf5_plugin_path / 'libcompression.dylib'
    elif platform.system() == 'Windows':
        remote_lib = 'https://share.mxwbio.com/d/4742248b2e674a85be97/files/?p=%2FWindows%2Fcompression.dll&dl=1'
        local_lib = hdf5_plugin_path / 'compression.dll'

    if not force_download and local_lib.is_file():
        print(f'lib h5 compression for maxwell already already in {local_lib}')
        return

    dist = urlopen(remote_lib)
    with open(local_lib, 'wb') as f:
        f.write(dist.read())



import unittest
from pathlib import Path
from numpy.testing import assert_array_equal, assert_

from neo.rawio.tdtrawio import TdtRawIO
from neo.test.rawiotest.common_rawio_test import BaseTestRawIO


class TestTdtRawIO(BaseTestRawIO, unittest.TestCase, ):
    rawioclass = TdtRawIO
    entities_to_download = [
        'tdt'
    ]
    entities_to_test = [
        'tdt/aep_05',
        'tdt/aep_05/Block-1/aep_05_Block-1.Tdx'
    ]

    def test_invalid_dirname(self):
        invalid_name = 'random_non_existant_tdt_filename'
        assert not Path(invalid_name).exists()

        with self.assertRaises(ValueError):
            TdtRawIO(invalid_name)

    def test_compare_load_multi_single_block(self):
        dirname = self.get_local_path('tdt/aep_05')
        filename = self.get_local_path('tdt/aep_05/Block-1/aep_05_Block-1.Tdx')

        io_single = TdtRawIO(filename)
        io_multi = TdtRawIO(dirname)

        io_single.parse_header()
        io_multi.parse_header()

        self.assertEqual(io_single.tdt_block_mode, 'single')
        self.assertEqual(io_multi.tdt_block_mode, 'multi')

        self.assertEqual(io_single.block_count(), 1)
        self.assertEqual(io_multi.block_count(), 1)

        self.assertEqual(io_single.segment_count(0), 1)
        self.assertEqual(io_multi.segment_count(0), 2)

        # compare header infos
        assert_array_equal(io_single.header['signal_streams'], io_multi.header['signal_streams'])
        assert_array_equal(io_single.header['signal_channels'], io_multi.header['signal_channels'])
        assert_array_equal(io_single.header['event_channels'], io_multi.header['event_channels'])

        # not all spiking channels are present in first tdt block (segment)
        for spike_channel in io_single.header['spike_channels']:
            self.assertIn(spike_channel, io_multi.header['spike_channels'])

        # check that extracted signal chunks are identical
        assert_array_equal(io_single.get_analogsignal_chunk(0, 0, 0, 100, 0),
                           io_multi.get_analogsignal_chunk(0, 0, 0, 100, 0))


if __name__ == "__main__":
    unittest.main()



"""
Tests of neo.rawio.mearecrawio

"""

import unittest

from neo.rawio.mearecrawio import MEArecRawIO

from neo.test.rawiotest.common_rawio_test import BaseTestRawIO


try:
    import MEArec as mr
    HAVE_MEAREC = True
except ImportError:
    HAVE_MEAREC = False


@unittest.skipUnless(HAVE_MEAREC, "requires MEArec package")
class TestMEArecRawIO(BaseTestRawIO, unittest.TestCase, ):
    rawioclass = MEArecRawIO
    entities_to_download = [
        'mearec'
    ]
    entities_to_test = [
        'mearec/mearec_test_10s.h5'
    ]


if __name__ == "__main__":
    unittest.main()



import unittest

from neo.io import RawMCSIO
from neo.test.iotest.common_io_test import BaseTestIO


class TestRawMcsIO(BaseTestIO, unittest.TestCase, ):
    ioclass = RawMCSIO
    entities_to_download = [
        'rawmcs'
    ]
    entities_to_test = [
        'rawmcs/raw_mcs_with_header_1.raw'
    ]


if __name__ == "__main__":
    unittest.main()



import math
import numpy as np


class NcsSections:
    """
    Contains information regarding the contiguous sections of records in an Ncs file.
    Methods of NcsSectionsFactory perform parsing of this information from an Ncs file and
    produce these where the sections are discontiguous in time and in temporal order.

    TODO: This class will likely need __ne__ to be useful in
    more sophisticated segment construction algorithms.

    """
    def __init__(self):
        self.sects = []
        self.sampFreqUsed = 0  # actual sampling frequency of samples
        self.microsPerSampUsed = 0  # microseconds per sample

    def __eq__(self, other):
        samp_eq = self.sampFreqUsed == other.sampFreqUsed
        micros_eq = self.microsPerSampUsed == other.microsPerSampUsed
        sects_eq = self.sects == other.sects
        return (samp_eq and micros_eq and sects_eq)

    def __hash__(self):
        return (f'{self.sampFreqUsed};{self.microsPerSampUsed};'
               f'{[s.__hash__() for s in self.sects]}').__hash__()


class NcsSection:
    """
    Information regarding a single contiguous section or group of records in an Ncs file.

    Model is that times are closed on the left and open on the right. Record
    numbers are closed on both left and right, that is, inclusive of the last record.

    endTime should never be set less than startTime for comparison functions to work
    properly, though this is not enforced.
    """

    _RECORD_SIZE = 512  # nb sample per signal record

    def __init__(self):
        self.startRec = -1  # index of starting record
        self.startTime = -1  # starttime of first record
        self.endRec = -1  # index of last record (inclusive)
        self.endTime = -1   # end time of last record, that is, the end time of the last
                            # sampling period contained in the last record of the section

    def __init__(self, sb, st, eb, et, ns):
        self.startRec = sb
        self.startTime = st
        self.endRec = eb
        self.endTime = et
        self.n_samples = ns

    def __eq__(self, other):
        return (self.startRec == other.startRec
                and self.startTime == other.startTime
                and self.endRec == other.endRec
                and self.endTime == other.endTime
                and self.n_samples == other.n_samples)

    def __hash__(self):
        s = f'{self.startRec};{self.startTime};{self.endRec};{self.endTime};{self.n_samples}'
        return s.__hash__()

    def before_time(self, rhb):
        """
        Determine if this section is completely before another section in time.
        """
        return self.endTime < rhb.startTime

    def overlaps_time(self, rhb):
        """
        Determine if this section overlaps another in time.
        """
        return self.startTime <= rhb.endTime and self.endTime >= rhb.startTime

    def after_time(self, rhb):
        """
        Determine if this section is completely after another section in time.
        """
        return self.startTime >= rhb.endTime


class NcsSectionsFactory:
    """
    Class for factory methods which perform parsing of contiguous sections of records
    in Ncs files.

    Model for times is that times are rounded to nearest microsecond. Times
    from start of a sample until just before the next sample are included,
    that is, closed lower bound and open upper bound on intervals. A
    channel with no samples is empty and contains no time intervals.

    Moved here since algorithm covering all 3 header styles and types used is
    more complicated.
    """

    _maxGapSampFrac = 0.2  # maximum fraction of a sampling interval between predicted
                           # and actual record timestamps still considered within one section

    @staticmethod
    def get_freq_for_micros_per_samp(micros):
        """
        Compute fractional sampling frequency, given microseconds per sample.
        """
        return 1e6 / micros

    @staticmethod
    def get_micros_per_samp_for_freq(sampFr):
        """
        Calculate fractional microseconds per sample, given the sampling frequency (Hz).
        """
        return 1e6 / sampFr

    @staticmethod
    def calc_sample_time(sampFr, startTime, posn):
        """
        Calculate time rounded to microseconds for sample given frequency,
        start time, and sample position.
        """
        return round(startTime + NcsSectionsFactory.get_micros_per_samp_for_freq(sampFr) * posn)

    @staticmethod
    def _parseGivenActualFrequency(ncsMemMap, ncsSects, chanNum, reqFreq, blkOnePredTime):
        """
        Parse sections in memory mapped file when microsPerSampUsed and sampFreqUsed are known,
        filling in an NcsSections object.

        PARAMETERS
        ncsMemMap:
            memmap of Ncs file
        ncsSections:
            NcsSections with actual sampFreqUsed correct, first NcsSection with proper startSect
            and startTime already added.
        chanNum:
            channel number that should be present in all records
        reqFreq:
            rounded frequency that all records should contain
        blkOnePredTime:
            predicted starting time of second record in block

        RETURN
        NcsSections object with block locations marked
        """
        startBlockPredTime = blkOnePredTime
        blk_len = 0
        curBlock = ncsSects.sects[0]
        for recn in range(1, ncsMemMap.shape[0]):
            timestamp = ncsMemMap['timestamp'][recn]
            channel_id = ncsMemMap['channel_id'][recn]
            sample_rate = ncsMemMap['sample_rate'][recn]
            nb_valid = ncsMemMap['nb_valid'][recn]

            if channel_id != chanNum or sample_rate != reqFreq:
                raise IOError('Channel number or sampling frequency changed in ' +
                              'records within file')
            predTime = NcsSectionsFactory.calc_sample_time(ncsSects.sampFreqUsed,
                                                           startBlockPredTime, blk_len)
            nValidSamps = nb_valid
            if timestamp != predTime:
                curBlock.endRec = recn - 1
                curBlock.endTime = predTime
                curBlock.n_samples = blk_len
                curBlock = NcsSection(recn, timestamp, -1, -1, -1)
                ncsSects.sects.append(curBlock)
                startBlockPredTime = NcsSectionsFactory.calc_sample_time(
                    ncsSects.sampFreqUsed,
                    timestamp,
                    nValidSamps)
                blk_len = 0
            else:
                blk_len += nValidSamps

        curBlock.endRec = ncsMemMap.shape[0] - 1
        endTime = NcsSectionsFactory.calc_sample_time(ncsSects.sampFreqUsed,
                                                      startBlockPredTime,
                                                      blk_len)
        curBlock.endTime = endTime

        return ncsSects

    @staticmethod
    def _buildGivenActualFrequency(ncsMemMap, actualSampFreq, reqFreq):
        """
        Build NcsSections object for file given actual sampling frequency.

        Requires that frequency in each record agrees with requested frequency. This is
        normally obtained by rounding the header frequency; however, this value may be different
        from the rounded actual frequency used in the recording, since the underlying
        requirement in older Ncs files was that the number of microseconds per sample in the
        records is the inverse of the sampling frequency stated in the header truncated to
        whole microseconds.

        PARAMETERS
        ncsMemMap:
            memmap of Ncs file
        actualSampFreq:
            actual sampling frequency used
        reqFreq:
            frequency to require in records

        RETURN:
            NcsSections object
        """
        # check frequency in first record
        if ncsMemMap['sample_rate'][0] != reqFreq:
            raise IOError("Sampling frequency in first record doesn't agree with header.")
        chanNum = ncsMemMap['channel_id'][0]

        nb = NcsSections()
        nb.sampFreqUsed = actualSampFreq
        nb.microsPerSampUsed = NcsSectionsFactory.get_micros_per_samp_for_freq(actualSampFreq)

        # check if file is one block of records, which is often the case, and avoid full parse
        lastBlkI = ncsMemMap.shape[0] - 1
        ts0 = ncsMemMap['timestamp'][0]
        nb0 = ncsMemMap['nb_valid'][0]
        predLastBlockStartTime = NcsSectionsFactory.calc_sample_time(actualSampFreq, ts0,
                                                                     NcsSection._RECORD_SIZE *
                                                                     lastBlkI)
        lts = ncsMemMap['timestamp'][lastBlkI]
        lnb = ncsMemMap['nb_valid'][lastBlkI]
        if ncsMemMap['channel_id'][lastBlkI] == chanNum and \
                ncsMemMap['sample_rate'][lastBlkI] == reqFreq and \
                lts == predLastBlockStartTime:
            lastBlkEndTime = NcsSectionsFactory.calc_sample_time(actualSampFreq, lts, lnb)
            n_samples = NcsSection._RECORD_SIZE * lastBlkI
            curBlock = NcsSection(0, ts0, lastBlkI, lastBlkEndTime, n_samples)

            nb.sects.append(curBlock)
            return nb

        # otherwise need to scan looking for breaks
        else:
            blkOnePredTime = NcsSectionsFactory.calc_sample_time(actualSampFreq, ts0, nb0)
            curBlock = NcsSection(0, ts0, -1, -1, -1)
            nb.sects.append(curBlock)
            return NcsSectionsFactory._parseGivenActualFrequency(ncsMemMap, nb, chanNum, reqFreq,
                                                                 blkOnePredTime)

    @staticmethod
    def _parseForMaxGap(ncsMemMap, ncsSects, maxGapLen):
        """
        Parse blocks of records from file, allowing a maximum gap in timestamps between records
        in sections. Estimates frequency being used based on timestamps.

        PARAMETERS
        ncsMemMap:
            memmap of Ncs file
        ncsSects:
            NcsSections object with sampFreqUsed set to nominal frequency to use in computing time
            for samples (Hz)
        maxGapLen:
            maximum difference within a block between predicted time of start of record and
            recorded time

        RETURN:
            NcsSections object with sampFreqUsed and microsPerSamp set based on estimate from
            largest block
        """

        chanNum = ncsMemMap['channel_id'][0]
        recFreq = ncsMemMap['sample_rate'][0]

        # check for consistent channel_ids and sampling rates
        ncsMemMap['channel_id']
        if not (ncsMemMap['channel_id'] == chanNum).all():
            raise IOError('Channel number changed in records within file')

        if not all(ncsMemMap['sample_rate'] == recFreq):
            raise IOError('Sampling frequency changed in records within file')

        # find most frequent number of samples
        exp_nb_valid = np.argmax(np.bincount(ncsMemMap['nb_valid']))
        # detect records with incomplete number of samples
        gap_rec_ids = list(np.where(ncsMemMap['nb_valid'] != exp_nb_valid)[0])

        rec_duration = 1e6 / ncsSects.sampFreqUsed * ncsMemMap['nb_valid']
        pred_times = np.rint(ncsMemMap['timestamp'] + rec_duration).astype(np.int64)
        max_pred_times = pred_times + maxGapLen
        # data records that start later than the predicted time (including the
        # maximal accepted gap length) are considered delayed and a gap is
        # registered.
        delayed_recs = list(np.where(max_pred_times[:-1] < ncsMemMap['timestamp'][1:])[0])
        gap_rec_ids.extend(delayed_recs)

        # cleaning extracted gap ids
        # last record can not be the beginning of a gap
        last_rec_id = len(ncsMemMap['timestamp']) - 1
        if last_rec_id in gap_rec_ids:
            gap_rec_ids.remove(last_rec_id)

        # gap ids can only be listed once
        gap_rec_ids = sorted(set(gap_rec_ids))

        # create recording segments from identified gaps
        ncsSects.sects.append(NcsSection(0, ncsMemMap['timestamp'][0], -1, -1, -1))
        for gap_rec_id in gap_rec_ids:
            curr_sec = ncsSects.sects[-1]
            curr_sec.endRec = gap_rec_id
            curr_sec.endTime = pred_times[gap_rec_id]
            n_samples = np.sum(ncsMemMap['nb_valid'][curr_sec.startRec:gap_rec_id + 1])
            curr_sec.n_samples = n_samples

            next_sec = NcsSection(gap_rec_id + 1,
                                  ncsMemMap['timestamp'][gap_rec_id + 1], -1, -1, -1)
            ncsSects.sects.append(next_sec)

        curr_sec = ncsSects.sects[-1]
        curr_sec.endRec = len(ncsMemMap['timestamp']) - 1
        curr_sec.endTime = pred_times[-1]
        n_samples = np.sum(ncsMemMap['nb_valid'][curr_sec.startRec:])
        curr_sec.n_samples = n_samples

        # calculate the estimated frequency of the block with the most samples
        max_blk_idx = np.argmax([bl.endRec - bl.startRec for bl in ncsSects.sects])
        max_blk = ncsSects.sects[max_blk_idx]

        maxBlkFreqEstimate = (max_blk.n_samples - ncsMemMap['nb_valid'][max_blk.endRec]) * 1e6 / \
                             (ncsMemMap['timestamp'][max_blk.endRec] - max_blk.startTime)

        ncsSects.sampFreqUsed = maxBlkFreqEstimate
        ncsSects.microsPerSampUsed = NcsSectionsFactory.get_micros_per_samp_for_freq(
                                                                        maxBlkFreqEstimate)
        # free memory that is unnecessarily occupied by the memmap
        # (see https://github.com/numpy/numpy/issues/19340)
        del ncsMemMap
        return ncsSects

    @staticmethod
    def _buildForMaxGap(ncsMemMap, nomFreq):
        """
        Determine sections of records in memory mapped Ncs file given a nominal frequency of
        the file, using the default values of frequency tolerance and maximum gap between blocks.

        PARAMETERS
        ncsMemMap:
            memmap of Ncs file
        nomFreq:
            nominal sampling frequency used, normally from header of file

        RETURN:
            NcsSections object
        """
        nb = NcsSections()

        numRecs = ncsMemMap.shape[0]
        if numRecs < 1:
            return nb

        chanNum = ncsMemMap['channel_id'][0]
        ts0 = ncsMemMap['timestamp'][0]

        lastBlkI = numRecs - 1
        lts = ncsMemMap['timestamp'][lastBlkI]
        lcid = ncsMemMap['channel_id'][lastBlkI]
        lnb = ncsMemMap['nb_valid'][lastBlkI]
        lsr = ncsMemMap['sample_rate'][lastBlkI]

        # check if file is one block of records, with exact timestamp match, which may be the case
        numSampsForPred = NcsSection._RECORD_SIZE * lastBlkI
        predLastBlockStartTime = NcsSectionsFactory.calc_sample_time(nomFreq, ts0, numSampsForPred)
        freqInFile = math.floor(nomFreq)
        if lts - predLastBlockStartTime == 0 and lcid == chanNum and lsr == freqInFile:
            endTime = NcsSectionsFactory.calc_sample_time(nomFreq, lts, lnb)
            curBlock = NcsSection(0, ts0, lastBlkI, endTime, numSampsForPred)
            nb.sects.append(curBlock)
            nb.sampFreqUsed = numSampsForPred / (lts - ts0) * 1e6
            nb.microsPerSampUsed = NcsSectionsFactory.get_micros_per_samp_for_freq(nb.sampFreqUsed)

        # otherwise parse records to determine blocks using default maximum gap length
        else:
            nb.sampFreqUsed = nomFreq
            nb.microsPerSampUsed = NcsSectionsFactory.get_micros_per_samp_for_freq(nb.sampFreqUsed)
            maxGapToAllow = round(NcsSectionsFactory._maxGapSampFrac * 1e6 / nomFreq)
            nb = NcsSectionsFactory._parseForMaxGap(ncsMemMap, nb, maxGapToAllow)

        return nb

    @staticmethod
    def build_for_ncs_file(ncsMemMap, nlxHdr):
        """
        Build an NcsSections object for an NcsFile, given as a memmap and NlxHeader,
        handling gap detection appropriately given the file type as specified by the header.

        PARAMETERS
        ncsMemMap:
            memory map of file
        nlxHdr:
            NlxHeader from corresponding file.

        RETURNS
        An NcsSections corresponding to the provided ncsMemMap and nlxHdr
        """
        acqType = nlxHdr.type_of_recording()

        # Old Neuralynx style with truncated whole microseconds for actual sampling. This
        # restriction arose from the sampling being based on a master 1 MHz clock.
        if acqType == "PRE4":
            freq = nlxHdr['sampling_rate']
            microsPerSampUsed = math.floor(NcsSectionsFactory.get_micros_per_samp_for_freq(freq))
            sampFreqUsed = NcsSectionsFactory.get_freq_for_micros_per_samp(microsPerSampUsed)
            nb = NcsSectionsFactory._buildGivenActualFrequency(ncsMemMap, sampFreqUsed,
                                                               math.floor(freq))
            nb.sampFreqUsed = sampFreqUsed
            nb.microsPerSampUsed = microsPerSampUsed

        # digital lynx style with fractional frequency and micros per samp determined from
        # block times
        elif acqType == "DIGITALLYNX" or acqType == "DIGITALLYNXSX":
            nomFreq = nlxHdr['sampling_rate']
            nb = NcsSectionsFactory._buildForMaxGap(ncsMemMap, nomFreq)

        # BML style with fractional frequency and micros per samp
        elif acqType == "BML":
            sampFreqUsed = nlxHdr['sampling_rate']
            nb = NcsSectionsFactory._buildGivenActualFrequency(ncsMemMap, sampFreqUsed,
                                                               math.floor(sampFreqUsed))

        else:
            raise TypeError("Unknown Ncs file type from header.")

        return nb

    @staticmethod
    def _verifySectionsStructure(ncsMemMap, ncsSects):
        """
        Check that the record structure and timestamps for the ncsMemMap
        agrees with that in ncsSects.

        Provides a more rapid verification of structure than building a new NcsSections
        and checking equality.

        PARAMETERS
        ncsMemMap:
            memmap of file to be checked
        ncsSects
            existing block structure to be checked
        RETURN:
            true if all timestamps and block record starts and stops agree, otherwise false.
        """
        for blki in range(0, len(ncsSects.sects)):
            if ncsMemMap['timestamp'][ncsSects.sects[blki].startRec] != \
                    ncsSects.sects[blki].startTime:
                return False
            ets = ncsMemMap['timestamp'][ncsSects.sects[blki].endRec]
            enb = ncsMemMap['nb_valid'][ncsSects.sects[blki].endRec]
            endTime = NcsSectionsFactory.calc_sample_time(ncsSects.sampFreqUsed, ets, enb)
            if endTime != ncsSects.sects[blki].endTime:
                return False

        return True



import unittest

from neo.io import CedIO
from neo.test.iotest.common_io_test import BaseTestIO


class TestCedIO(BaseTestIO, unittest.TestCase, ):
    ioclass = CedIO
    entities_to_test = [
        'spike2/m365_1sec.smrx',
        'spike2/File_spike2_1.smr',
        'spike2/Two-mice-bigfile-test000.smr'
    ]
    entities_to_download = [
        'spike2'
    ]


if __name__ == "__main__":
    unittest.main()



"""
Tests of neo.io.neomatlabio
"""

import unittest

from neo.io import MicromedIO
from neo.test.iotest.common_io_test import BaseTestIO


class TestMicromedIO(BaseTestIO, unittest.TestCase, ):
    ioclass = MicromedIO
    entities_to_download = [
        'micromed'
    ]
    entities_to_test = [
        'micromed/File_micromed_1.TRC'
    ]


if __name__ == "__main__":
    unittest.main()



"""
baserawio
======

Classes
-------

BaseRawIO
abstract class which should be overridden to write a RawIO.

RawIO is a low level API in neo that provides fast  access to the raw data.
When possible, all IOs should/implement this level following these guidelines:
  * internal use of memmap (or hdf5)
  * fast reading of the header (do not read the complete file)
  * neo tree object is symmetric and logical: same channel/units/event
    along all block and segments.

For this level, datasets of recordings are mapped as follows:

A channel refers to a physical channel of recording in an experiment. It is identified by a
channel_id. Recordings from a channel consist of sections of samples which are recorded
contiguously in time; in other words, a section of a channel has a specific sampling_rate,
start_time, and length (and thus also stop_time, which is the time of the sample which would
lie one sampling interval beyond the last sample present in that section).

A stream consists of a set of channels which all have the same structure of their sections of
recording and the same data type of samples. Each stream has a unique stream_id and has a name,
which does not need to be unique. A stream thus has multiple channels which all have the same
sampling rate and are on the same clock, have the same sections with t_starts and lengths, and
the same data type for their samples. The samples in a stream can thus be retrieved as an Numpy
array, a chunk of samples.

Channels within a stream can be accessed be either their channel_id, which must be unique within
a stream, or by their channel_index, which is a 0 based index to all channels within the stream.
Note that a single channel of recording may be represented within multiple streams, and such is
the case for RawIOs which may have both unfiltered and filtered or downsampled versions of the
signals from a single recording channel. In such a case, a single channel and channel_id may be
represented by a different channel_index within different streams. Lists of channel_indexes are
often convenient to pass around a selection of channels within a stream.

At the neo.io level, one AnalogSignal with multiple channels can be created for each stream. Such
an AnalogSignal may have multiple Segments, with each segment containing the sections from each
channel with the same t_start and length. Such multiple Segments for a RawIO will have the
same sampling rate. It is thus possible to retrieve the t_start and length
of the sections of the channels for a Block and Segment of a stream.

So this handles **only** one simplified but very frequent case of dataset:
    * Only one channel set  for AnalogSignal stable along Segment
    * Only one channel set  for SpikeTrain stable along Segment
    * AnalogSignal have all the same sampling_rate acroos all Segment
    * t_start/t_stop are the same for many object (SpikeTrain, Event) inside a Segment

Signal channels are handled by group of "stream".
One stream will result at neo.io level in one AnalogSignal with multiple channels.

A helper class `neo.io.basefromrawio.BaseFromRaw` transforms a RawIO to
neo legacy IO. In short all "neo.rawio" classes are also "neo.io"
with lazy reading capability.

With this API the IO have an attributes `header` with necessary keys.
This  `header` attribute is done in `_parse_header(...)` method.
See ExampleRawIO as example.

BaseRawIO also implements a possible persistent cache system that can be used
by some RawIOs to avoid a very long parse_header() call. The idea is that some variable
or vector can be stored somewhere (near the file, /tmp, any path) for use across multiple
constructions of a RawIO for a given set of data.

"""

import logging
import numpy as np
import os
import sys

from neo import logging_handler

try:
    import joblib

    HAVE_JOBLIB = True
except ImportError:
    HAVE_JOBLIB = False

possible_raw_modes = ['one-file', 'multi-file', 'one-dir', ]  # 'multi-dir', 'url', 'other'

error_header = 'Header is not read yet, do parse_header() first'

_signal_stream_dtype = [
    ('name', 'U64'),  # not necessarily unique
    ('id', 'U64'),  # must be unique
]

_signal_channel_dtype = [
    ('name', 'U64'),  # not necessarily unique
    ('id', 'U64'),  # must be unique
    ('sampling_rate', 'float64'),
    ('dtype', 'U16'),
    ('units', 'U64'),
    ('gain', 'float64'),
    ('offset', 'float64'),
    ('stream_id', 'U64'),
]

# TODO for later: add t_start and length in _signal_channel_dtype
# this would simplify all t_start/t_stop stuff for each RawIO class

_common_sig_characteristics = ['sampling_rate', 'dtype', 'stream_id']

_spike_channel_dtype = [
    ('name', 'U64'),
    ('id', 'U64'),
    # for waveform
    ('wf_units', 'U64'),
    ('wf_gain', 'float64'),
    ('wf_offset', 'float64'),
    ('wf_left_sweep', 'int64'),
    ('wf_sampling_rate', 'float64'),
]

# in rawio event and epoch are handled the same way
# except, that duration is `None` for events
_event_channel_dtype = [
    ('name', 'U64'),
    ('id', 'U64'),
    ('type', 'S5'),  # epoch or event
]


class BaseRawIO:
    """
    Generic class to handle.

    """

    name = 'BaseIO'
    description = ''
    extensions = []

    rawmode = None  # one key from possible_raw_modes

    def __init__(self, use_cache=False, cache_path='same_as_resource', **kargs):
        """
        :TODO: Why multi-file would have a single filename is confusing here - shouldn't
        the name of this argument be filenames_list or filenames_base or similar?

        When rawmode=='one-file' kargs MUST contains 'filename' the filename
        When rawmode=='multi-file' kargs MUST contains 'filename' one of the filenames.
        When rawmode=='one-dir' kargs MUST contains 'dirname' the dirname.

        """
        # create a logger for the IO class
        fullname = self.__class__.__module__ + '.' + self.__class__.__name__
        self.logger = logging.getLogger(fullname)
        # Create a logger for 'neo' and add a handler to it if it doesn't have one already.
        # (it will also not add one if the root logger has a handler)
        corename = self.__class__.__module__.split('.')[0]
        corelogger = logging.getLogger(corename)
        rootlogger = logging.getLogger()
        if not corelogger.handlers and not rootlogger.handlers:
            corelogger.addHandler(logging_handler)

        self.use_cache = use_cache
        if use_cache:
            assert HAVE_JOBLIB, 'You need to install joblib for cache'
            self.setup_cache(cache_path)
        else:
            self._cache = None

        self.header = None

    def parse_header(self):
        """
        This must parse the file header to get all stuff for fast use later on.

        This must create
        self.header['nb_block']
        self.header['nb_segment']
        self.header['signal_streams']
        self.header['signal_channels']
        self.header['spike_channels']
        self.header['event_channels']

        """
        self._parse_header()
        self._check_stream_signal_channel_characteristics()

    def source_name(self):
        """Return fancy name of file source"""
        return self._source_name()

    def __repr__(self):
        txt = '{}: {}\n'.format(self.__class__.__name__, self.source_name())
        if self.header is not None:
            nb_block = self.block_count()
            txt += 'nb_block: {}\n'.format(nb_block)
            nb_seg = [self.segment_count(i) for i in range(nb_block)]
            txt += 'nb_segment:  {}\n'.format(nb_seg)

            # signal streams
            v = [s['name'] + f' (chans: {self.signal_channels_count(i)})'
                for i, s in enumerate(self.header['signal_streams'])]
            v = pprint_vector(v)
            txt += f'signal_streams: {v}\n'

            for k in ('signal_channels', 'spike_channels', 'event_channels'):
                ch = self.header[k]
                v = pprint_vector(self.header[k]['name'])
                txt += f'{k}: {v}\n'

        return txt

    def _generate_minimal_annotations(self):
        """
        Helper function that generates a nested dict for annotations.

        Must be called when these are Ok after self.header is done
        and thus when these functions return the correct values:
          * block_count()
          * segment_count()
          * signal_streams_count()
          * signal_channels_count()
          * spike_channels_count()
          * event_channels_count()

        There are several sources and kinds of annotations that will
        be forwarded to the neo.io level and used to enrich neo objects:
            * annotations of objects common across segments
                * signal_streams > neo.AnalogSignal annotations
                * signal_channels > neo.AnalogSignal array_annotations split by stream
                * spike_channels > neo.SpikeTrain
                * event_channels > neo.Event and neo.Epoch
            * annotations that depend of the block_id/segment_id of the object:
              * nested in raw_annotations['blocks'][block_index]['segments'][seg_index]['signals']

        Usage after a call to this function we can do this to populate more annotations:

        raw_annotations['blocks'][block_index][ 'nickname'] = 'super block'
        raw_annotations['blocks'][block_index]
                        ['segments']['important_key'] = 'important value'
        raw_annotations['blocks'][block_index]
                        ['segments'][seg_index]
                        ['signals']['nickname'] = 'super signals stream'
        raw_annotations['blocks'][block_index]
                        ['segments'][seg_index]
                        ['signals']['__array_annotations__']
                        ['channels_quality'] = ['bad', 'good', 'medium', 'good']
        raw_annotations['blocks'][block_index]
                        ['segments'][seg_index]
                        ['spikes'][spike_chan]['nickname'] =  'super neuron'
        raw_annotations['blocks'][block_index]
                        ['segments'][seg_index]
                        ['spikes'][spike_chan]
                        ['__array_annotations__']['spike_amplitudes'] = [-1.2, -10., ...]
        raw_annotations['blocks'][block_index]
                        ['segments'][seg_index]
                        ['events'][ev_chan]['nickname'] = 'super trigger'
        raw_annotations['blocks'][block_index]
                        ['segments'][seg_index]
                        ['events'][ev_chan]
                        Z['__array_annotations__']['additional_label'] = ['A', 'B', 'A', 'C', ...]


        Theses annotations will be used at the neo.io API directly in objects.

        Standard annotation like name/id/file_origin are already generated here.
        """
        signal_streams = self.header['signal_streams']
        signal_channels = self.header['signal_channels']
        spike_channels = self.header['spike_channels']
        event_channels = self.header['event_channels']

        # use for AnalogSignal.annotations and AnalogSignal.array_annotations
        signal_stream_annotations = []
        for c in range(signal_streams.size):
            stream_id = signal_streams[c]['id']
            channels = signal_channels[signal_channels['stream_id'] == stream_id]
            d = {}
            d['name'] = signal_streams['name'][c]
            d['stream_id'] = stream_id
            d['file_origin'] = self._source_name()
            d['__array_annotations__'] = {}
            for key in ('name', 'id'):
                values = np.array([channels[key][chan] for chan in range(channels.size)])
                d['__array_annotations__']['channel_' + key + 's'] = values
            signal_stream_annotations.append(d)

        # used for SpikeTrain.annotations and SpikeTrain.array_annotations
        spike_annotations = []
        for c in range(spike_channels.size):
            # use for Unit.annotations
            d = {}
            d['name'] = spike_channels['name'][c]
            d['id'] = spike_channels['id'][c]
            d['file_origin'] = self._source_name()
            d['__array_annotations__'] = {}
            spike_annotations.append(d)

        # used for Event/Epoch.annotations and Event/Epoch.array_annotations
        event_annotations = []
        for c in range(event_channels.size):
            # not used in neo.io at the moment could useful one day
            d = {}
            d['name'] = event_channels['name'][c]
            d['id'] = event_channels['id'][c]
            d['file_origin'] = self._source_name()
            d['__array_annotations__'] = {}
            event_annotations.append(d)

        # duplicate this signal_stream_annotations/spike_annotations/event_annotations
        # across blocks and segments and create annotations
        ann = {}
        ann['blocks'] = []
        for block_index in range(self.block_count()):
            d = {}
            d['file_origin'] = self.source_name()
            d['segments'] = []
            ann['blocks'].append(d)

            for seg_index in range(self.segment_count(block_index)):
                d = {}
                d['file_origin'] = self.source_name()
                # copy nested
                d['signals'] = signal_stream_annotations.copy()
                d['spikes'] = spike_annotations.copy()
                d['events'] = event_annotations.copy()
                ann['blocks'][block_index]['segments'].append(d)

        self.raw_annotations = ann

    def _repr_annotations(self):
        txt = 'Raw annotations\n'
        for block_index in range(self.block_count()):
            bl_a = self.raw_annotations['blocks'][block_index]
            txt += '*Block {}\n'.format(block_index)
            for k, v in bl_a.items():
                if k in ('segments',):
                    continue
                txt += '  -{}: {}\n'.format(k, v)
            for seg_index in range(self.segment_count(block_index)):
                seg_a = bl_a['segments'][seg_index]
                txt += '  *Segment {}\n'.format(seg_index)
                for k, v in seg_a.items():
                    if k in ('signals', 'spikes', 'events',):
                        continue
                    txt += '    -{}: {}\n'.format(k, v)

                # annotations by channels for spikes/events/epochs
                for child in ('signals', 'events', 'spikes', ):
                    if child == 'signals':
                        n = self.header['signal_streams'].shape[0]
                    else:
                        n = self.header[child[:-1] + '_channels'].shape[0]
                    for c in range(n):
                        neo_name = {'signals': 'AnalogSignal',
                                    'spikes': 'SpikeTrain',
                                    'events': 'Event/Epoch'}[child]
                        txt += f'    *{neo_name} {c}\n'
                        child_a = seg_a[child][c]
                        for k, v in child_a.items():
                            if k == '__array_annotations__':
                                continue
                            txt += f'      -{k}: {v}\n'
                        for k, values in child_a['__array_annotations__'].items():
                            values = ', '.join([str(v) for v in values[:4]])
                            values = '[ ' + values + ' ...'
                            txt += f'      -{k}: {values}\n'

        return txt

    def print_annotations(self):
        """Print formatted raw_annotations"""
        print(self._repr_annotations())

    def block_count(self):
        """return number of blocks"""
        return self.header['nb_block']

    def segment_count(self, block_index):
        """return number of segments for a given block"""
        return self.header['nb_segment'][block_index]

    def signal_streams_count(self):
        """Return the number of signal streams.
        Same for all Blocks and Segments.
        """
        return len(self.header['signal_streams'])

    def signal_channels_count(self, stream_index):
        """Return the number of signal channels for a given stream.
        This number is the same for all Blocks and Segments.
        """
        stream_id = self.header['signal_streams'][stream_index]['id']
        channels = self.header['signal_channels']
        channels = channels[channels['stream_id'] == stream_id]
        return len(channels)

    def spike_channels_count(self):
        """Return the number of unit (aka spike) channels.
        Same for all Blocks and Segments.
        """
        return len(self.header['spike_channels'])

    def event_channels_count(self):
        """Return the number of event/epoch channels.
        Same for all Blocks and Segments.
        """
        return len(self.header['event_channels'])

    def segment_t_start(self, block_index, seg_index):
        """Global t_start of a Segment in s. Shared by all objects except
        for AnalogSignal.
        """
        return self._segment_t_start(block_index, seg_index)

    def segment_t_stop(self, block_index, seg_index):
        """Global t_start of a Segment in s. Shared by all objects except
        for AnalogSignal.
        """
        return self._segment_t_stop(block_index, seg_index)

    ###
    # signal and channel zone

    def _check_stream_signal_channel_characteristics(self):
        """
        Check that all channels that belonging to the same stream_id
        have the same stream id and _common_sig_characteristics. These
        presently include:
          * sampling_rate
          * units
          * dtype
        """
        signal_streams = self.header['signal_streams']
        signal_channels = self.header['signal_channels']
        if signal_streams.size > 0:
            assert signal_channels.size > 0, 'Signal stream but no signal_channels!!!'

        for stream_index in range(signal_streams.size):
            stream_id = signal_streams[stream_index]['id']
            mask = signal_channels['stream_id'] == stream_id
            characteristics = signal_channels[mask][_common_sig_characteristics]
            unique_characteristics = np.unique(characteristics)
            assert unique_characteristics.size == 1, \
                f'Some channel in stream_id {stream_id} ' \
                f'do not have same {_common_sig_characteristics} {unique_characteristics}'

            # also check that channel_id is unique inside a stream
            channel_ids = signal_channels[mask]['id']
            assert np.unique(channel_ids).size == channel_ids.size, \
                f'signal_channels dont have unique ids for stream {stream_index}'

        self._several_channel_groups = signal_streams.size > 1

    def channel_name_to_index(self, stream_index, channel_names):
        """
        Inside a stream, transform channel_names to channel_indexes.
        Based on self.header['signal_channels']
        channel_indexes are zero-based offsets within the stream
        """
        stream_id = self.header['signal_streams'][stream_index]['id']
        mask = self.header['signal_channels']['stream_id'] == stream_id
        signal_channels = self.header['signal_channels'][mask]
        chan_names = list(signal_channels['name'])
        assert signal_channels.size == np.unique(chan_names).size, 'Channel names not unique'
        channel_indexes = np.array([chan_names.index(name) for name in channel_names])
        return channel_indexes

    def channel_id_to_index(self, stream_index, channel_ids):
        """
        Inside a stream, transform channel_ids to channel_indexes.
        Based on self.header['signal_channels']
        channel_indexes are zero-based offsets within the stream
        """
        # unique ids is already checked in _check_stream_signal_channel_characteristics
        stream_id = self.header['signal_streams'][stream_index]['id']
        mask = self.header['signal_channels']['stream_id'] == stream_id
        signal_channels = self.header['signal_channels'][mask]
        chan_ids = list(signal_channels['id'])
        channel_indexes = np.array([chan_ids.index(chan_id) for chan_id in channel_ids])
        return channel_indexes

    def _get_channel_indexes(self, stream_index, channel_indexes, channel_names, channel_ids):
        """
        Select channel_indexes for a stream based on channel_indexes/channel_names/channel_ids
        depending which is not None.
        """
        if channel_indexes is None and channel_names is not None:
            channel_indexes = self.channel_name_to_index(stream_index, channel_names)
        elif channel_indexes is None and channel_ids is not None:
            channel_indexes = self.channel_id_to_index(stream_index, channel_ids)
        return channel_indexes

    def _get_stream_index_from_arg(self, stream_index_arg):
        if stream_index_arg is None:
            assert self.header['signal_streams'].size == 1
            stream_index = 0
        else:
            assert 0 <= stream_index_arg < self.header['signal_streams'].size
            stream_index = stream_index_arg
        return stream_index

    def get_signal_size(self, block_index, seg_index, stream_index=None):
        """
        Retrieve the length of a single section of the channels in a stream.
        :param block_index:
        :param seg_index:
        :param stream_index:
        :return: number of samples
        """
        stream_index = self._get_stream_index_from_arg(stream_index)
        return self._get_signal_size(block_index, seg_index, stream_index)

    def get_signal_t_start(self, block_index, seg_index, stream_index=None):
        """
        Retrieve the t_start of a single section of the channels in a stream.
        :param block_index:
        :param seg_index:
        :param stream_index:
        :return: start time of section
        """
        stream_index = self._get_stream_index_from_arg(stream_index)
        return self._get_signal_t_start(block_index, seg_index, stream_index)

    def get_signal_sampling_rate(self, stream_index=None):
        """
        Retrieve sampling rate for a stream and all channels in that stream.
        :param stream_index:
        :return: sampling rate
        """
        stream_index = self._get_stream_index_from_arg(stream_index)
        stream_id = self.header['signal_streams'][stream_index]['id']
        mask = self.header['signal_channels']['stream_id'] == stream_id
        signal_channels = self.header['signal_channels'][mask]
        sr = signal_channels[0]['sampling_rate']
        return float(sr)

    def get_analogsignal_chunk(self, block_index=0, seg_index=0, i_start=None, i_stop=None,
                               stream_index=None, channel_indexes=None, channel_names=None,
                               channel_ids=None, prefer_slice=False):
        """
        Return a chunk of raw signal as a Numpy array. columns correspond to samples from a
        section of a single channel of recording. The channels are chosen either by channel_names,
        if provided, otherwise by channel_ids, if provided, otherwise by channel_indexes, if
        provided, otherwise all channels are selected.

        :param block_index: block containing segment with section
        :param seg_index: segment containing section
        :param i_start: index of first sample to retrieve within section
        :param i_stop: index of one past last sample to retrieve within section
        :param stream_index: index of stream containing channels
        :param channel_indexes: list of indexes of channels to retrieve. Can be a list, slice,
                                  np.array of int, or None
        :param channel_names: list of channels names to retrieve, or None
        :param channel_ids: list of channel ids to retrieve, or None
        :param prefer_slice: use slicing with lazy read if channel_indexes are provided as an
                              np.ndarray and are contiguous
        :return: array with raw signal samples
        """
        stream_index = self._get_stream_index_from_arg(stream_index)
        channel_indexes = self._get_channel_indexes(stream_index, channel_indexes,
                                                    channel_names, channel_ids)

        # some check on channel_indexes
        if isinstance(channel_indexes, list):
            channel_indexes = np.asarray(channel_indexes)

        if isinstance(channel_indexes, np.ndarray):
            if channel_indexes.dtype == 'bool':
                assert self.signal_channels_count(stream_index) == channel_indexes.size
                channel_indexes, = np.nonzero(channel_indexes)

        if prefer_slice and isinstance(channel_indexes, np.ndarray):
            # Check if channel_indexes are contiguous and transform to slice argument if possible.
            # This is useful for memmap or hdf5 where providing a slice causes a lazy read,
            # rather than a list of indexes that make a copy (like numpy.take()).
            if np.all(np.diff(channel_indexes) == 1):
                channel_indexes = slice(channel_indexes[0], channel_indexes[-1] + 1)

        raw_chunk = self._get_analogsignal_chunk(
            block_index, seg_index, i_start, i_stop, stream_index, channel_indexes)

        return raw_chunk

    def rescale_signal_raw_to_float(self, raw_signal, dtype='float32', stream_index=None,
                                    channel_indexes=None, channel_names=None, channel_ids=None):
        """
        Rescale a chunk of raw signals which are provided as a Numpy array. These are normally
        returned by a call to get_analog_signal_chunk. The channels are specified either by
        channel_names, if provided, otherwise by channel_ids, if provided, otherwise by
        channel_indexes, if provided, otherwise all channels are selected.

        :param raw_signal: Numpy array of samples. columns are samples for a single channel
        :param dtype: data type for returned scaled samples
        :param stream_index: index of stream containing channels
        :param channel_indexes: list of indexes of channels to retrieve or None
        :param channel_names: list of channels names to retrieve, or None
        :param channel_ids: list of channel ids to retrieve, or None
        :return: array of scaled sample values
        """
        stream_index = self._get_stream_index_from_arg(stream_index)
        channel_indexes = self._get_channel_indexes(stream_index, channel_indexes,
                                                    channel_names, channel_ids)
        if channel_indexes is None:
            channel_indexes = slice(None)

        stream_id = self.header['signal_streams'][stream_index]['id']
        mask = self.header['signal_channels']['stream_id'] == stream_id
        channels = self.header['signal_channels'][mask]
        if channel_indexes is None:
            channel_indexes = slice(None)
        channels = channels[channel_indexes]

        float_signal = raw_signal.astype(dtype)

        if np.any(channels['gain'] != 1.):
            float_signal *= channels['gain']

        if np.any(channels['offset'] != 0.):
            float_signal += channels['offset']

        return float_signal

    # spiketrain and unit zone
    def spike_count(self, block_index=0, seg_index=0, spike_channel_index=0):
        return self._spike_count(block_index, seg_index, spike_channel_index)

    def get_spike_timestamps(self, block_index=0, seg_index=0, spike_channel_index=0,
                             t_start=None, t_stop=None):
        """
        The timestamp datatype is as close to the format itself. Sometimes float/int32/int64.
        Sometimes it is the index on the signal but not always.
        The conversion to second or index_on_signal is done outside this method.

        t_start/t_stop are limits in seconds.
        """
        timestamp = self._get_spike_timestamps(block_index, seg_index,
                                               spike_channel_index, t_start, t_stop)
        return timestamp

    def rescale_spike_timestamp(self, spike_timestamps, dtype='float64'):
        """
        Rescale spike timestamps to seconds.
        """
        return self._rescale_spike_timestamp(spike_timestamps, dtype)

    # spiketrain waveform zone
    def get_spike_raw_waveforms(self, block_index=0, seg_index=0, spike_channel_index=0,
                                t_start=None, t_stop=None):
        wf = self._get_spike_raw_waveforms(block_index, seg_index,
                                           spike_channel_index, t_start, t_stop)
        return wf

    def rescale_waveforms_to_float(self, raw_waveforms, dtype='float32',
                                   spike_channel_index=0):
        wf_gain = self.header['spike_channels']['wf_gain'][spike_channel_index]
        wf_offset = self.header['spike_channels']['wf_offset'][spike_channel_index]

        float_waveforms = raw_waveforms.astype(dtype)

        if wf_gain != 1.:
            float_waveforms *= wf_gain
        if wf_offset != 0.:
            float_waveforms += wf_offset

        return float_waveforms

    # event and epoch zone
    def event_count(self, block_index=0, seg_index=0, event_channel_index=0):
        return self._event_count(block_index, seg_index, event_channel_index)

    def get_event_timestamps(self, block_index=0, seg_index=0, event_channel_index=0,
                             t_start=None, t_stop=None):
        """
        The timestamp datatype is as close to the format itself. Sometimes float/int32/int64.
        Sometimes it is the index on the signal but not always.
        The conversion to second or index_on_signal is done outside this method.

        t_start/t_sop are limits in seconds.

        returns
            timestamp
            labels
            durations

        """
        timestamp, durations, labels = self._get_event_timestamps(
            block_index, seg_index, event_channel_index, t_start, t_stop)
        return timestamp, durations, labels

    def rescale_event_timestamp(self, event_timestamps, dtype='float64',
                    event_channel_index=0):
        """
        Rescale event timestamps to seconds.
        """
        return self._rescale_event_timestamp(event_timestamps, dtype, event_channel_index)

    def rescale_epoch_duration(self, raw_duration, dtype='float64',
                    event_channel_index=0):
        """
        Rescale epoch raw duration to seconds.
        """
        return self._rescale_epoch_duration(raw_duration, dtype, event_channel_index)

    def setup_cache(self, cache_path, **init_kargs):
        if self.rawmode in ('one-file', 'multi-file'):
            resource_name = self.filename
        elif self.rawmode == 'one-dir':
            resource_name = self.dirname
        else:
            raise (NotImplementedError)

        if cache_path == 'home':
            if sys.platform.startswith('win'):
                dirname = os.path.join(os.environ['APPDATA'], 'neo_rawio_cache')
            elif sys.platform.startswith('darwin'):
                dirname = '~/Library/Application Support/neo_rawio_cache'
            else:
                dirname = os.path.expanduser('~/.config/neo_rawio_cache')
            dirname = os.path.join(dirname, self.__class__.__name__)

            if not os.path.exists(dirname):
                os.makedirs(dirname)
        elif cache_path == 'same_as_resource':
            dirname = os.path.dirname(resource_name)
        else:
            assert os.path.exists(cache_path), \
                'cache_path do not exists use "home" or "same_as_resource" to make this auto'

        # the hash of the resource (dir of file) is done with filename+datetime
        # TODO make something more sophisticated when rawmode='one-dir' that use all
        #  filename and datetime
        d = dict(ressource_name=resource_name, mtime=os.path.getmtime(resource_name))
        hash = joblib.hash(d, hash_name='md5')

        # name is constructed from the real_n,ame and the hash
        name = '{}_{}'.format(os.path.basename(resource_name), hash)
        self.cache_filename = os.path.join(dirname, name)

        if os.path.exists(self.cache_filename):
            self.logger.warning('Use existing cache file {}'.format(self.cache_filename))
            self._cache = joblib.load(self.cache_filename)
        else:
            self.logger.warning('Create cache file {}'.format(self.cache_filename))
            self._cache = {}
            self.dump_cache()

    def add_in_cache(self, **kargs):
        assert self.use_cache
        self._cache.update(kargs)
        self.dump_cache()

    def dump_cache(self):
        assert self.use_cache
        joblib.dump(self._cache, self.cache_filename)

    ##################

    # Functions to be implemented in IO below here

    def _parse_header(self):
        raise (NotImplementedError)
        # must call
        # self._generate_empty_annotations()

    def _source_name(self):
        raise (NotImplementedError)

    def _segment_t_start(self, block_index, seg_index):
        raise (NotImplementedError)

    def _segment_t_stop(self, block_index, seg_index):
        raise (NotImplementedError)

    ###
    # signal and channel zone
    def _get_signal_size(self, block_index, seg_index, stream_index):
        """
        Return the size of a set of AnalogSignals indexed by channel_indexes.

        All channels indexed must have the same size and t_start.
        """
        raise (NotImplementedError)

    def _get_signal_t_start(self, block_index, seg_index, stream_index):
        """
        Return the t_start of a set of AnalogSignals indexed by channel_indexes.

        All channels indexed must have the same size and t_start.
        """
        raise (NotImplementedError)

    def _get_analogsignal_chunk(self, block_index, seg_index, i_start, i_stop,
                                stream_index, channel_indexes):
        """
        Return the samples from a set of AnalogSignals indexed
        by stream_index and channel_indexes (local index inner stream).

        RETURNS
        -------
            array of samples, with each requested channel in a column
        """

        raise (NotImplementedError)

    ###
    # spiketrain and unit zone
    def _spike_count(self, block_index, seg_index, spike_channel_index):
        raise (NotImplementedError)

    def _get_spike_timestamps(self, block_index, seg_index,
                              spike_channel_index, t_start, t_stop):
        raise (NotImplementedError)

    def _rescale_spike_timestamp(self, spike_timestamps, dtype):
        raise (NotImplementedError)

    ###
    # spike waveforms zone
    def _get_spike_raw_waveforms(self, block_index, seg_index,
                                 spike_channel_index, t_start, t_stop):
        raise (NotImplementedError)

    ###
    # event and epoch zone
    def _event_count(self, block_index, seg_index, event_channel_index):
        raise (NotImplementedError)

    def _get_event_timestamps(self, block_index, seg_index, event_channel_index, t_start, t_stop):
        raise (NotImplementedError)

    def _rescale_event_timestamp(self, event_timestamps, dtype):
        raise (NotImplementedError)

    def _rescale_epoch_duration(self, raw_duration, dtype):
        raise (NotImplementedError)


def pprint_vector(vector, lim=8):
    vector = np.asarray(vector)
    assert vector.ndim == 1
    if len(vector) > lim:
        part1 = ', '.join(e for e in vector[:lim // 2])
        part2 = ' , '.join(e for e in vector[-lim // 2:])
        txt = f"[{part1} ... {part2}]"
    else:
        part1 = ', '.join(e for e in vector[:lim // 2])
        txt = f"[{part1}]"
    return txt



"""
Tests of neo.io.wineedrio
"""

import unittest

from neo.io import WinEdrIO
from neo.test.iotest.common_io_test import BaseTestIO


class TestWinedrIO(BaseTestIO, unittest.TestCase, ):
    ioclass = WinEdrIO
    entities_to_download = [
        'winedr'
    ]
    entities_to_test = [
        'winedr/File_WinEDR_1.EDR',
        'winedr/File_WinEDR_2.EDR',
        'winedr/File_WinEDR_3.EDR',
    ]


if __name__ == "__main__":
    unittest.main()



import unittest

from neo.rawio.neuroscoperawio import NeuroScopeRawIO
from neo.test.rawiotest.common_rawio_test import BaseTestRawIO


class TestNeuroScopeRawIO(BaseTestRawIO, unittest.TestCase, ):
    rawioclass = NeuroScopeRawIO
    entities_to_download = [
        'neuroscope'
    ]
    entities_to_test = [
        'neuroscope/test1/test1'
    ]


if __name__ == "__main__":
    unittest.main()



import unittest

from neo.rawio.spikegadgetsrawio import SpikeGadgetsRawIO
from neo.test.rawiotest.common_rawio_test import BaseTestRawIO


class TestSpikeGadgetsRawIO(BaseTestRawIO, unittest.TestCase, ):
    rawioclass = SpikeGadgetsRawIO
    entities_to_download = ['spikegadgets']
    entities_to_test = [
        'spikegadgets/20210225_em8_minirec2_ac.rec',
        'spikegadgets/W122_06_09_2019_1_fromSD.rec'
    ]


if __name__ == "__main__":
    unittest.main()



import unittest

from neo.io import SpikeGadgetsIO
from neo.test.iotest.common_io_test import BaseTestIO


class TestSpikeGadgetsIO(BaseTestIO, unittest.TestCase, ):
    ioclass = SpikeGadgetsIO
    entities_to_download = ['spikegadgets']
    entities_to_test = [
        'spikegadgets/20210225_em8_minirec2_ac.rec',
        'spikegadgets/W122_06_09_2019_1_fromSD.rec'
    ]


if __name__ == "__main__":
    unittest.main()



"""
Class for reading spikegadgets files.
Only continuous signals are supported at the moment.

https://spikegadgets.com/spike-products/

Documentation of the format:
https://bitbucket.org/mkarlsso/trodes/wiki/Configuration

Note :
  * this file format have multiple version. news version include the gain for scaling.
     The actual implementation do not contain this feature because we don't have
     files to test this. So now the gain is "hardcoded" to 1. and so units
     is not handled correctly.

The ".rec" file format contains:
  * a first  text part with information in an XML structure
  * a second part for the binary buffer

Author: Samuel Garcia
"""
from .baserawio import (BaseRawIO, _signal_channel_dtype, _signal_stream_dtype,
    _spike_channel_dtype, _event_channel_dtype)

import numpy as np

from xml.etree import ElementTree


class SpikeGadgetsRawIO(BaseRawIO):
    extensions = ['rec']
    rawmode = 'one-file'

    def __init__(self, filename='', selected_streams=None):
        """
        Class for reading spikegadgets files.
        Only continuous signals are supported at the moment.

        Initialize a SpikeGadgetsRawIO for a single ".rec" file.

        Args:
            filename: str
                The filename
            selected_streams: None, list, str
                sublist of streams to load/expose to API
                useful for spikeextractor when one stream only is needed.
                For instance streams = ['ECU', 'trodes']
                'trodes' is name for ephy channel (ntrodes)
        """
        BaseRawIO.__init__(self)
        self.filename = filename
        self.selected_streams = selected_streams

    def _source_name(self):
        return self.filename

    def _parse_header(self):
        # parse file until "</Configuration>"
        header_size = None
        with open(self.filename, mode='rb') as f:
            while True:
                line = f.readline()
                if b"</Configuration>" in line:
                    header_size = f.tell()
                    break

            if header_size is None:
                ValueError("SpikeGadgets: the xml header does not contain '</Configuration>'")

            f.seek(0)
            header_txt = f.read(header_size).decode('utf8')

        # explore xml header
        root = ElementTree.fromstring(header_txt)
        gconf = sr = root.find('GlobalConfiguration')
        hconf = root.find('HardwareConfiguration')
        sconf = root.find('SpikeConfiguration')

        self._sampling_rate = float(hconf.attrib['samplingRate'])
        num_ephy_channels = int(hconf .attrib['numChannels'])

        # explore sub stream and count packet size
        # first bytes is 0x55
        packet_size = 1
        stream_bytes = {}
        for device in hconf:
            stream_id = device.attrib['name']
            num_bytes = int(device.attrib['numBytes'])
            stream_bytes[stream_id] = packet_size
            packet_size += num_bytes

        # timesteamps 4 uint32
        self._timestamp_byte = packet_size
        packet_size += 4

        packet_size += 2 * num_ephy_channels

        # read the binary part lazily
        raw_memmap = np.memmap(self.filename, mode='r', offset=header_size, dtype='<u1')

        num_packet = raw_memmap.size // packet_size
        raw_memmap = raw_memmap[:num_packet * packet_size]
        self._raw_memmap = raw_memmap.reshape(-1, packet_size)

        # create signal channels
        stream_ids = []
        signal_streams = []
        signal_channels = []

        # walk in xml device and keep only "analog" one
        self._mask_channels_bytes = {}
        for device in hconf:
            stream_id = device.attrib['name']
            for channel in device:

                if 'interleavedDataIDByte' in channel.attrib:
                    # TODO LATER: deal with "headstageSensor" which have interleaved
                    continue

                if channel.attrib['dataType'] == 'analog':

                    if stream_id not in stream_ids:
                        stream_ids.append(stream_id)
                        stream_name = stream_id
                        signal_streams.append((stream_name, stream_id))
                        self._mask_channels_bytes[stream_id] = []

                    name = channel.attrib['id']
                    chan_id = channel.attrib['id']
                    dtype = 'int16'
                    # TODO LATER : handle gain correctly according the file version
                    units = ''
                    gain = 1.
                    offset = 0.
                    signal_channels.append((name, chan_id, self._sampling_rate, 'int16',
                                         units, gain, offset, stream_id))

                    num_bytes = stream_bytes[stream_id] + int(channel.attrib['startByte'])
                    chan_mask = np.zeros(packet_size, dtype='bool')
                    chan_mask[num_bytes] = True
                    chan_mask[num_bytes + 1] = True
                    self._mask_channels_bytes[stream_id].append(chan_mask)

        if num_ephy_channels > 0:
            stream_id = 'trodes'
            stream_name = stream_id
            signal_streams.append((stream_name, stream_id))
            self._mask_channels_bytes[stream_id] = []

            chan_ind = 0
            for trode in sconf:
                for schan in trode:
                    name = 'trode' + trode.attrib['id'] + 'chan' + schan.attrib['hwChan']
                    chan_id = schan.attrib['hwChan']
                    # TODO LATER : handle gain correctly according the file version
                    units = ''
                    gain = 1.
                    offset = 0.
                    signal_channels.append((name, chan_id, self._sampling_rate, 'int16',
                                         units, gain, offset, stream_id))

                    chan_mask = np.zeros(packet_size, dtype='bool')
                    num_bytes = packet_size - 2 * num_ephy_channels + 2 * chan_ind
                    chan_mask[num_bytes] = True
                    chan_mask[num_bytes + 1] = True
                    self._mask_channels_bytes[stream_id].append(chan_mask)

                    chan_ind += 1

        # make mask as array (used in _get_analogsignal_chunk(...))
        self._mask_streams = {}
        for stream_id, l in self._mask_channels_bytes.items():
            mask = np.array(l)
            self._mask_channels_bytes[stream_id] = mask
            self._mask_streams[stream_id] = np.any(mask, axis=0)

        signal_streams = np.array(signal_streams, dtype=_signal_stream_dtype)
        signal_channels = np.array(signal_channels, dtype=_signal_channel_dtype)

        # remove some stream if no wanted
        if self.selected_streams is not None:
            if isinstance(self.selected_streams, str):
                self.selected_streams = [self.selected_streams]
            assert isinstance(self.selected_streams, list)

            keep = np.in1d(signal_streams['id'], self.selected_streams)
            signal_streams = signal_streams[keep]

            keep = np.in1d(signal_channels['stream_id'], self.selected_streams)
            signal_channels = signal_channels[keep]

        # No events channels
        event_channels = []
        event_channels = np.array(event_channels, dtype=_event_channel_dtype)

        # No spikes  channels
        spike_channels = []
        spike_channels = np.array(spike_channels, dtype=_spike_channel_dtype)

        # fille into header dict
        self.header = {}
        self.header['nb_block'] = 1
        self.header['nb_segment'] = [1]
        self.header['signal_streams'] = signal_streams
        self.header['signal_channels'] = signal_channels
        self.header['spike_channels'] = spike_channels
        self.header['event_channels'] = event_channels

        self._generate_minimal_annotations()
        # info from GlobalConfiguration in xml are copied to block and seg annotations
        bl_ann = self.raw_annotations['blocks'][0]
        seg_ann = self.raw_annotations['blocks'][0]['segments'][0]
        for ann in (bl_ann, seg_ann):
            ann.update(gconf.attrib)

    def _segment_t_start(self, block_index, seg_index):
        return 0.

    def _segment_t_stop(self, block_index, seg_index):
        size = self._raw_memmap.shape[0]
        t_stop = size / self._sampling_rate
        return t_stop

    def _get_signal_size(self, block_index, seg_index, stream_index):
        size = self._raw_memmap.shape[0]
        return size

    def _get_signal_t_start(self, block_index, seg_index, stream_index):
        return 0.

    def _get_analogsignal_chunk(self, block_index, seg_index, i_start, i_stop, stream_index,
                                channel_indexes):
        stream_id = self.header['signal_streams'][stream_index]['id']

        raw_unit8 = self._raw_memmap[i_start:i_stop]

        num_chan = len(self._mask_channels_bytes[stream_id])
        re_order = None
        if channel_indexes is None:
            # no loop : entire stream mask
            stream_mask = self._mask_streams[stream_id]
        else:
            # accumulate mask
            if isinstance(channel_indexes, slice):
                chan_inds = np.arange(num_chan)[channel_indexes]
            else:
                chan_inds = channel_indexes

                if np.any(np.diff(channel_indexes) < 0):
                    # handle channel are not ordered
                    sorted_channel_indexes = np.sort(channel_indexes)
                    re_order = np.array([list(sorted_channel_indexes).index(ch)
                                         for ch in channel_indexes])

            stream_mask = np.zeros(raw_unit8.shape[1], dtype='bool')
            for chan_ind in chan_inds:
                chan_mask = self._mask_channels_bytes[stream_id][chan_ind]
                stream_mask |= chan_mask

        # this copies the data from the memmap into memory
        raw_unit8_mask = raw_unit8[:, stream_mask]
        shape = raw_unit8_mask.shape
        shape = (shape[0], shape[1] // 2)
        # reshape the and retype by view
        raw_unit16 = raw_unit8_mask.flatten().view('int16').reshape(shape)

        if re_order is not None:
            raw_unit16 = raw_unit16[:, re_order]

        return raw_unit16



import unittest
import os

from neo.io import MaxwellIO
from neo.test.iotest.common_io_test import BaseTestIO

from neo.rawio.maxwellrawio import auto_install_maxwell_hdf5_compression_plugin


class TestMaxwellIO(BaseTestIO, unittest.TestCase, ):
    ioclass = MaxwellIO
    entities_to_download = [
        'maxwell'
    ]
    entities_to_test = files_to_test = [
        'maxwell/MaxOne_data/Record/000011/data.raw.h5',
        'maxwell/MaxTwo_data/Network/000028/data.raw.h5'
    ]

    def setUp(self):
        auto_install_maxwell_hdf5_compression_plugin(force_download=False)
        BaseTestIO.setUp(self)


if __name__ == "__main__":
    unittest.main()



import unittest

from neo.rawio.winedrrawio import WinEdrRawIO
from neo.test.rawiotest.common_rawio_test import BaseTestRawIO


class TestWinEdrRawIO(BaseTestRawIO, unittest.TestCase, ):
    rawioclass = WinEdrRawIO
    entities_to_download = [
        'winedr'
    ]
    entities_to_test = [
        'winedr/File_WinEDR_1.EDR',
        'winedr/File_WinEDR_2.EDR',
        'winedr/File_WinEDR_3.EDR',
    ]


if __name__ == "__main__":
    unittest.main()
